# Copyright 2023 The PECOS Developers
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance with
# the License.You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
# specific language governing permissions and limitations under the License.

"""Quantum circuit to PHIR conversion utilities.

This module provides functions to convert PECOS QuantumCircuit objects to
PHIR (PECOS High-level Intermediate Representation) format for execution
on quantum hardware and classical simulators.
"""

from __future__ import annotations

import json
from copy import deepcopy
from typing import TYPE_CHECKING, Any

import pecos

if TYPE_CHECKING:
    from collections.abc import Sequence

# TODO: Does this handle parallel gate ticks?

qsym_conv = {
    "init |0>": "Init",
    "measure Z": "Measure",
    "SqrtX": "SX",
    "SqrtXd": "SXdg",
    "SqrtY": "SY",
    "SqrtYd": "SYdg",
    "SqrtZ": "SZ",
    "SqrtZd": "SZdg",
    "SqrtZZ": "SZZ",
}


def conv_expr(expr: dict[str, Any]) -> dict[str, Any]:
    """Convert expression dictionary to PHIR format.

    Transforms expression dictionaries from quantum circuit format to PHIR
    (PECOS High-level Intermediate Representation) format, handling assignment
    operations and binary operations with proper argument structure.

    Args:
        expr: Expression dictionary with operation and operand information.

    Returns:
        Converted expression dictionary in PHIR format with 'cop', 'args',
        and optionally 'returns' fields.
    """
    if "t" in expr:
        op = "="

        if expr["op"] == "=":
            # op = = -> "t = a"
            if "b" in expr:
                msg = "Assignment expression should not have 'b' operand"
                raise ValueError(msg)
            left = expr["t"]
            right = expr["a"]
        else:
            # op != = -> "t = a op b"
            left = expr["t"]
            right = {"op": expr["op"], "a": expr["a"], "b": expr["b"]}

    else:
        op = expr["op"]
        left = expr["a"]
        right = expr.get("b")

    if isinstance(left, dict):
        left = conv_expr(left)

    if isinstance(right, dict):
        right = conv_expr(right)

    new_expr = {
        "cop": op,
    }

    if isinstance(left, tuple):
        left = list(left)

    if isinstance(right, tuple):
        right = list(right)

    if op == "=":
        new_expr["args"] = [right]
        new_expr["returns"] = [left]
    elif right is not None:
        new_expr["args"] = [left, right]
    else:
        new_expr["args"] = [left]

    return new_expr


def to_phir_dict(qc: pecos.QuantumCircuit) -> dict:
    """Creates a json str representation of the QuantumCircuit listing all the gates.

    It does not preserve ticks or parallel gating of different gate types.
    """
    prog = {
        "format": "PHIR/JSON",
        "version": "0.1.0",
        "metadata": deepcopy(qc.metadata),
        "ops": [],
    }
    prog["metadata"]["source_program_type"] = [
        "PECOS.QuantumCircuit",
        ["PECOS", str(pecos.__version__)],
    ]

    data = {"max_qid": 0}
    if "qvars" in qc.metadata:
        qid2qsym = {k: list(v) for k, v in prog["metadata"]["qvars"].items()}
        del prog["metadata"]["qvars"]
        prog["metadata"]["qid2qubit_sym"] = qid2qsym
    else:
        qid2qsym = {}

    def get_qsym(qid: int) -> list[str | int]:
        qsym = qid2qsym.get(qid)

        if qsym is None:
            if "qvar_spec" not in qc.metadata:
                qsym = ["q", data["max_qid"]]
                data["max_qid"] += 1
                qid2qsym[qid] = qsym
            else:
                msg = f"Couldn't find qsym for qid = {qid}"
                raise Exception(msg)
        return list(qsym)

    def find_qid2qsym(
        qubits: Sequence[int | tuple[int, ...] | list[int]],
    ) -> list[str | list[str]]:
        qs = []
        for loc in qubits:
            if isinstance(loc, tuple | list):
                qtup = []
                for q in loc:
                    qsym = get_qsym(q)

                    qtup.append(qsym)
                qs.append(qtup)
            else:
                qsym = get_qsym(loc)
                qs.append(qsym)
        return qs

    def flush_ops_buffer(
        ops: list[dict[str, Any]],
        ops_buffer: list[dict[str, Any]],
        current_cond: dict[str, Any] | None,
    ) -> None:
        if current_cond is None:
            ops.extend(ops_buffer)
        else:
            block = {
                "block": "if",
                "condition": current_cond,
                "true_branch": ops_buffer,
            }
            ops.append(block)

    def handle_ops_buffer(
        current_op: dict[str, Any] | None,
        ops: list[dict[str, Any]],
        ops_buffer: list[dict[str, Any]],
        cond: dict[str, Any] | None,
        current_cond: dict[str, Any] | None,
        *,
        force_flush: bool = False,
    ) -> tuple[list[dict[str, Any]], list[dict[str, Any]], dict[str, Any] | None]:
        if (cond != current_cond or force_flush) and ops_buffer:
            flush_ops_buffer(ops, ops_buffer, current_cond)
            ops_buffer = []
            current_cond = cond

        if current_op:
            ops_buffer = ops_buffer_append(ops_buffer, current_op)

        return ops, ops_buffer, current_cond

    def ops_buffer_append(
        ops_buffer: list[dict[str, Any]],
        current_op: dict[str, Any],
        *,
        compact_gates: bool = True,
    ) -> list[dict[str, Any]]:
        # TODO: When you do that make sure gates don't overlap

        if ops_buffer and compact_gates:
            last_op = ops_buffer[-1]
            if (
                ("qop" in last_op and "qop" in current_op and last_op["qop"] == current_op["qop"])
                or (last_op.get("data") == last_op.get("data") == "cvar_export")
            ) and last_op.get("metadata") == current_op.get("metadata"):
                if "args" in current_op:
                    # TODO: Check no overlap!
                    last_op["args"].extend(current_op["args"])

                if "variables" in current_op:
                    last_op["variables"].extend(current_op["variables"])

                if "returns" in current_op:
                    last_op["returns"].extend(current_op["returns"])
            else:
                ops_buffer.append(current_op)

        else:
            ops_buffer.append(current_op)

        return ops_buffer

    ops = prog["ops"]

    for sym, size in qc.metadata.get("qvar_spec", {}).items():
        ops.append(
            {
                "data": "qvar_define",
                "data_type": "qubits",
                "variable": sym,
                "size": size,
            },
        )

    for sym, size in qc.metadata.get("cvar_spec", {}).items():
        ops.append(
            {
                "data": "cvar_define",
                "data_type": "i32",
                "variable": sym,
                "size": size,
            },
        )

    ops_buffer = []
    current_cond = None

    for sym, qubits, meta in qc.items():
        op = {}
        metadata = deepcopy(meta) if meta else {}

        if sym == "cop":
            # TODO: HOW to identify mops...., need to tag at higher level
            if metadata.get("cop_type") == "CFunc":
                op = {
                    "cop": "ffcall",
                    "function": metadata["func"],
                    "args": metadata["args"],
                }

                if metadata["assign_vars"]:
                    op["returns"] = metadata["assign_vars"]

            elif metadata.get("cop_type") == "ExportCVar":
                del metadata["cop_type"]
                if "duration" in metadata:
                    del metadata["duration"]

                op = {
                    "data": "cvar_export",
                    "variables": [metadata["export"]],
                }
                del metadata["export"]

            elif metadata.get("cop_type") in {"Idle", "idle"}:
                del metadata["cop_type"]
                op = {
                    "mop": "Idle",
                    "args": find_qid2qsym(qubits),
                }

            elif metadata.get("cop_type") in {"Transport", "transport"}:
                del metadata["cop_type"]
                op = {
                    "mop": "Transport",
                    "args": find_qid2qsym(qubits),
                }

            elif "expr" in metadata:
                op = conv_expr(metadata["expr"])
                del metadata["expr"]

            else:
                print("!!!!!", qubits, metadata)

        else:  # qop
            op.update(
                {
                    "qop": qsym_conv.get(sym, sym),
                    "args": find_qid2qsym(qubits),
                },
            )

            angles = None
            if "angle" in metadata:
                angles = [metadata["angle"]]
            elif "angles" in metadata:
                angles = metadata["angles"]

            if angles:
                op["angles"] = angles

            if sym.startswith("measure"):
                # Getting return values:
                if "var" in metadata:
                    var = metadata["var"]
                    op["returns"] = [list(var)]
                    del metadata["var"]
                elif "var_output" in metadata:
                    var_output = metadata["var_output"]
                    var_return = []
                    for q in qubits:
                        var = var_output[q]
                        if isinstance(var, str):
                            var_return.append(var)
                        else:
                            var_return.append(list(var))
                    op["returns"] = var_return
                    del metadata["var_output"]

        cond = metadata.get("cond")
        if cond is not None:
            cond = conv_expr(cond)

        if "cond" in metadata:
            del metadata["cond"]

        if metadata:
            op["metadata"] = metadata

        ops, ops_buffer, current_cond = handle_ops_buffer(
            op,
            ops,
            ops_buffer,
            cond,
            current_cond,
        )

    # Flush the buffer of any remaining operations
    if ops_buffer:
        ops, ops_buffer, current_cond = handle_ops_buffer(
            None,
            ops,
            ops_buffer,
            None,
            current_cond,
            force_flush=True,
        )

    num_qubits = len(qid2qsym)
    prog["metadata"]["num_qubits"] = num_qubits

    if "qvar_spec" not in qc.metadata:
        op = {
            "data": "qvar_define",
            "data_type": "qubits",
            "variable": "q",
            "size": num_qubits,
        }

        ops.insert(0, op)

    if "cvar_spec" in prog["metadata"]:
        del prog["metadata"]["cvar_spec"]

    if "qvar_spec" in prog["metadata"]:
        del prog["metadata"]["qvar_spec"]

    return prog


def to_phir_json(qc: pecos.QuantumCircuit) -> str:
    """Convert the QuantumCircuit to the PHIR/JSON format."""
    return json.dumps(to_phir_dict(qc))



# Copyright 2018 The PECOS Developers
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance with
# the License.You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
# specific language governing permissions and limitations under the License.

"""Quantum measurement operations for Pauli fault propagation simulator.

This module provides quantum measurement operations for the Pauli fault propagation simulator, including projective
measurements with efficient Pauli frame updates and fault propagation tracking through measurement operations.
"""

from __future__ import annotations

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from pecos.simulators.pauliprop.state import PauliProp
    from pecos.typing import SimulatorGateParams


def meas_x(state: PauliProp, qubit: int, **_params: SimulatorGateParams) -> int:
    """Measurement in the X basis.

    Args:
        state: The PauliProp state instance.
        qubit (int): The qubit index to measure.
        **_params: Unused additional parameters (kept for interface compatibility).
    """
    if qubit in state.faults["Z"] or qubit in state.faults["Y"]:
        return 1
    return 0


def meas_z(state: PauliProp, qubit: int, **_params: SimulatorGateParams) -> int:
    """Measurement in the Z basis.

    Args:
        state: The PauliProp state instance.
        qubit (int): The qubit index to measure.
        **_params: Unused additional parameters (kept for interface compatibility).
    """
    if qubit in state.faults["X"] or qubit in state.faults["Y"]:
        return 1
    return 0


def meas_y(state: PauliProp, qubit: int, **_params: SimulatorGateParams) -> int:
    """Measurement in the Y basis.

    Args:
        state: The PauliProp state instance.
        qubit (int): The qubit index to measure.
        **_params: Unused additional parameters (kept for interface compatibility).
    """
    if qubit in state.faults["X"] or qubit in state.faults["Z"]:
        return 1
    return 0


def meas_pauli(
    state: PauliProp,
    qubits: int | tuple[int, ...],
    **params: SimulatorGateParams,
) -> int:
    """Measure a multi-qubit Pauli operator.

    Performs measurement of a specified Pauli operator on one or more qubits,
    returning the parity of individual Pauli measurements on each qubit.

    Args:
        state: Pauli fault propagation state containing fault information.
        qubits: Qubit index or tuple of qubit indices to measure.
        **params: Gate parameters including 'Pauli' specifying the operator.

    Returns:
        Measurement result (0 or 1) as parity of individual measurements.

    Raises:
        Exception: If Pauli operator specification is invalid.
    """
    pauli = params["Pauli"]

    if isinstance(qubits, int) and pauli not in {"X", "Y", "Z"}:
        msg = "Pauli for a single qubit measurement must be 'X', 'Y' or 'Z'!"
        raise Exception(msg)

    if pauli in {"X", "Y", "Z"}:
        pauli *= len(qubits)
    elif len(pauli) == len(qubits) + 1:
        # last qubit is considered the syndrome ancilla
        qubits = qubits[:-1]
    elif len(pauli) != len(qubits):
        msg = "The Pauli operator needs to be the size of the qubits it is acting on or a single type."
        raise Exception(msg)

    meas = 0

    for q, p in zip(qubits, pauli, strict=False):
        if p == "X":
            meas += meas_x(state, q)
        elif p == "Z":
            meas += meas_z(state, q)
        elif p == "Y":
            meas += meas_y(state, q)
        else:
            msg = "Pauli symbol not supported!"
            raise Exception(msg)

    return meas % 2


def force_output(_state: PauliProp, _qubit: int, forced_output: int = -1) -> int:
    """Outputs value.

    Used for error generators to generate outputs when replacing measurements.

    Args:
        _state: Unused state parameter (kept for interface compatibility).
        _qubit (int): Unused qubit parameter (kept for interface compatibility).
        forced_output (int): The value to output.
    """
    return forced_output



# Copyright 2025 The PECOS Developers
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance with
# the License.You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
# specific language governing permissions and limitations under the License.

"""Regression tests for logical Steane code teleportation programs."""

from collections.abc import Callable

from pecos.slr import Barrier, CReg, If, Main
from pecos.slr.qeclib.steane.steane_class import Steane


def telep(prep_basis: str, meas_basis: str) -> Main:
    """Generate teleportation program for logical Steane code."""
    return Main(
        m_bell := CReg("m_bell", size=2),
        m_out := CReg("m_out", size=1),
        # Input state:
        sin := Steane("sin", default_rus_limit=2),
        smid := Steane("smid"),
        sout := Steane("sout"),
        # Create Bell state
        smid.pz(),  # prep logical qubit in |0>/|+Z> state with repeat-until-success initialization
        sout.pz(),
        Barrier(smid.d, sout.d),
        smid.h(),
        smid.cx(sout),  # CX with control on smid and target on sout
        smid.qec(),
        sout.qec(),
        # prepare input state in some Pauli basis state
        sin.p(prep_basis, rus_limit=3),
        sin.qec(),
        # entangle input with one of the logical qubits of the Bell pair
        sin.cx(smid),
        sin.h(),
        # Bell measurement
        sin.mz(m_bell[0]),
        smid.mz(m_bell[1]),
        # Corrections
        If(m_bell[1] == 0).Then(sout.x()),
        If(m_bell[0] == 0).Then(sout.z()),
        # Final output stored in `m_out[0]`
        sout.m(meas_basis, m_out[0]),
    )


def test_telep(compare_qasm: Callable[..., None]) -> None:
    """Test teleportation regression with different prep and measurement bases."""
    for prep_basis in ["-X", "+X", "+Y", "-Y", "+Z", "-Z"]:
        for meas_basis in ["X", "Y", "Z"]:
            prog = telep(prep_basis, meas_basis)
            compare_qasm(
                prog,
                prep_basis,
                meas_basis,
                filename="local_steane_code_program.telep",
            )


def t_gate(prep_basis: str, meas_basis: str) -> Main:
    """A simple example of teleporting a T gate on a state.

    Args:
        prep_basis (str):  A string indicating what Pauli basis to prepare the state in. Acceptable inputs include:
            "+X", "-X", "+Y", "-Y", "+Z", and "-Z".
        meas_basis (str): A string indicating what Pauli basis the measure out the logical qubit in. Acceptable inputs
            include: "X", "Y", and "Z".

    Returns:
        A logical program written in extended OpenQASM 2.0
    """
    return Main(
        m_reject := CReg("m_reject", size=2),
        m_t := CReg("m_t", 1),
        m_out := CReg("m_out", size=2),
        # Input state:
        sin := Steane("sin", default_rus_limit=2),
        saux := Steane("saux"),
        sin.p(prep_basis, rus_limit=3),
        # Apply T to `sin` using the logical ancilla saux and indicating whether t prep failed.
        sin.t(saux, reject=m_reject[0], rus_limit=1),
        # Final output stored in `m_out[0]`
        sin.m(meas_basis, m_out[0]),
        # More explicitly, create the resource state and teleport the gate
        # ----------------------------------------------------------------
        # Prepare the resource state
        saux.prep_t_plus_state(reject=m_reject[1], rus_limit=2),
        # Prepare the input state
        sin.p(prep_basis),
        # teleport the T gate onto the input state
        sin.cx(saux),
        saux.mz(m_t[0]),
        If(sin.t_meas == 1).Then(sin.sz()),
        # Final output stored in `m_out[1]`
        sin.m(meas_basis, m_out[1]),
    )


def test_t_gate(compare_qasm: Callable[..., None]) -> None:
    """Test T-gate teleportation regression with different prep and measurement bases."""
    for prep_basis in ["-X", "+X", "+Y", "-Y", "+Z", "-Z"]:
        for meas_basis in ["X", "Y", "Z"]:
            prog = t_gate(prep_basis, meas_basis)
            compare_qasm(
                prog,
                prep_basis,
                meas_basis,
                filename="local_steane_code_program.t_gate",
            )



# Copyright 2026 The PECOS Developers
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance with
# the License. You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
# specific language governing permissions and limitations under the License.

"""Gate cancellation optimization pass.

Removes consecutive self-inverse gates acting on the same qubits.
For example: X-X, H-H, CX-CX (on same control/target) all cancel to identity.
"""

from __future__ import annotations

from typing import TYPE_CHECKING

from pecos.slr.ast.optimizations.base import StatementListOptimizer
from pecos.slr.ast.optimizations.gate_properties import is_self_inverse, targets_match

if TYPE_CHECKING:
    from pecos.slr.ast.nodes import GateOp


class GateCancellationPass(StatementListOptimizer):
    """Remove consecutive self-inverse gates acting on the same qubits.

    Self-inverse gates satisfy G * G = I, so consecutive identical gates
    on the same qubits can be removed entirely.

    Supported cancellations:
    - X-X, Y-Y, Z-Z (Pauli gates)
    - H-H (Hadamard)
    - CX-CX, CY-CY, CZ-CZ, CH-CH (two-qubit Clifford gates on same qubits)

    Example:
        # Before optimization
        X(q[0]), X(q[0])  # These cancel

        # After optimization
        (empty - both gates removed)
    """

    @property
    def name(self) -> str:
        return "gate_cancellation"

    def _should_cancel(self, gate1: GateOp, gate2: GateOp) -> bool:
        """Check if two gates cancel (same self-inverse gate on same qubits).

        Gates cancel if:
        1. They are the same gate type
        2. The gate type is self-inverse
        3. They act on the same qubits in the same order
        4. Neither has parameters (parameterized gates handled by rotation_merging)
        """
        return (
            gate1.gate == gate2.gate
            and is_self_inverse(gate1.gate)
            and targets_match(gate1, gate2)
            and not gate1.params
            and not gate2.params
        )



"""Test real quantum circuits through the Guppy->HUGR->Selene->ByteMessage pipeline."""

import pytest
from guppylang import guppy
from guppylang.std.angles import angle
from guppylang.std.quantum import cx, h, measure, qubit, ry, rz, x, z
from pecos import Guppy, sim
from pecos_rslib import state_vector

pytestmark = pytest.mark.optional_dependency


def test_bell_state_preparation() -> None:
    """Test Bell state preparation and measurement."""

    @guppy
    def prepare_bell_state() -> tuple[bool, bool]:
        """Prepare a Bell state |Φ+⟩ = (|00⟩ + |11⟩)/√2."""
        q1 = qubit()
        q2 = qubit()

        # Create Bell state
        h(q1)
        cx(q1, q2)

        # Measure both qubits
        m1 = measure(q1)
        m2 = measure(q2)

        return (m1, m2)

    # Run simulation with state_vector backend
    # Use seed for reproducibility
    shot_vec = sim(Guppy(prepare_bell_state)).qubits(2).quantum(state_vector()).seed(42).run(1000)
    assert shot_vec is not None, "Should get results"
    results = shot_vec.to_dict()
    # Count outcomes
    both_zero = 0
    both_one = 0
    anti_correlated = 0

    # Results come as a dict with measurement keys
    m1_list = results.get("measurement_0", [])
    m2_list = results.get("measurement_1", [])

    for m1, m2 in zip(m1_list, m2_list, strict=False):
        if m1 == 0 and m2 == 0:
            both_zero += 1
        elif m1 == 1 and m2 == 1:
            both_one += 1
        else:
            anti_correlated += 1

    # Bell state should only produce correlated outcomes
    assert anti_correlated == 0, f"Bell state should not produce anti-correlated outcomes, got {anti_correlated}"
    assert both_zero > 0, "Should see |00⟩ outcomes"
    assert both_one > 0, "Should see |11⟩ outcomes"

    # Should be roughly 50/50 split
    total = both_zero + both_one
    assert 0.4 < both_zero / total < 0.6, f"Should be ~50% |00⟩, got {both_zero / total}"
    assert 0.4 < both_one / total < 0.6, f"Should be ~50% |11⟩, got {both_one / total}"


def test_ghz_state() -> None:
    """Test 3-qubit GHZ state preparation."""

    @guppy
    def prepare_ghz_state() -> tuple[bool, bool, bool]:
        """Prepare a GHZ state |GHZ⟩ = (|000⟩ + |111⟩)/√2."""
        q1 = qubit()
        q2 = qubit()
        q3 = qubit()

        # Create GHZ state
        h(q1)
        cx(q1, q2)
        cx(q1, q3)

        # Measure all qubits
        m1 = measure(q1)
        m2 = measure(q2)
        m3 = measure(q3)

        return (m1, m2, m3)

    # Run simulation with state_vector backend
    shot_vec = sim(Guppy(prepare_ghz_state)).qubits(3).quantum(state_vector()).seed(42).run(1000)
    assert shot_vec is not None, "Should get results"
    results = shot_vec.to_dict()

    # GHZ state should give either all 0s or all 1s
    all_zero = 0
    all_one = 0
    other = 0

    m1_list = results.get("measurement_0", [])
    m2_list = results.get("measurement_1", [])
    m3_list = results.get("measurement_2", [])

    for m1, m2, m3 in zip(m1_list, m2_list, m3_list, strict=False):
        if m1 == 0 and m2 == 0 and m3 == 0:
            all_zero += 1
        elif m1 == 1 and m2 == 1 and m3 == 1:
            all_one += 1
        else:
            other += 1

    # GHZ state should only produce |000⟩ or |111⟩
    assert other == 0, f"GHZ state should not produce mixed outcomes, got {other}"
    assert all_zero > 0, "Should see |000⟩ outcomes"
    assert all_one > 0, "Should see |111⟩ outcomes"


def test_quantum_phase_kickback() -> None:
    """Test quantum phase kickback circuit."""

    @guppy
    def phase_kickback_circuit() -> tuple[bool, bool]:
        """Demonstrate phase kickback with controlled-Z gate."""
        control = qubit()
        target = qubit()

        # Put control in superposition
        h(control)

        # Put target in |1⟩ state
        x(target)

        # Apply controlled-Z (phase kickback occurs)
        # Since we don't have cz directly, use the equivalence: CZ = H·CX·H
        h(target)
        cx(control, target)
        h(target)

        # Measure in X basis for control (apply H before measure)
        h(control)
        m1 = measure(control)

        # Measure target in Z basis
        m2 = measure(target)

        return (m1, m2)

    # Run simulation with state_vector backend
    results = sim(Guppy(phase_kickback_circuit)).qubits(2).quantum(state_vector()).seed(42).run(1000)
    assert results is not None, "Should get results"

    # The control qubit should measure |1⟩ in X basis (due to phase kickback)
    # The target should remain in |1⟩
    control_one_count = 0
    target_one_count = 0
    total = 0

    if hasattr(results, "__getitem__"):
        m1_list = results.get("measurement_0", [])
        m2_list = results.get("measurement_1", [])

        for m1, m2 in zip(m1_list, m2_list, strict=False):
            total += 1
            if m1 == 1:
                control_one_count += 1
            if m2 == 1:
                target_one_count += 1

    # Control should be predominantly |1⟩ due to phase kickback
    assert (
        control_one_count / total > 0.9
    ), f"Control should be ~100% |1⟩ after phase kickback, got {control_one_count / total}"
    # Target should remain |1⟩
    assert target_one_count / total > 0.9, f"Target should remain |1⟩, got {target_one_count / total}"


def test_quantum_interference() -> None:
    """Test quantum interference in a simple interferometer."""

    @guppy
    def quantum_interferometer() -> bool:
        """Create quantum interference using H gates."""
        q = qubit()

        # First H gate - creates superposition
        h(q)

        # Phase shift of π
        z(q)

        # Second H gate - creates interference
        h(q)

        # Should measure |1⟩ due to destructive interference
        return measure(q)

    # Run simulation with state_vector backend
    results = sim(Guppy(quantum_interferometer)).qubits(1).quantum(state_vector()).seed(42).run(1000)
    assert results is not None, "Should get results"

    # Due to interference, should measure |1⟩ ~100% of the time
    one_count = 0
    total = 0

    if hasattr(results, "__getitem__"):
        measurements = results.get("measurement_0", [])
        for m in measurements:
            total += 1
            if m == 1:
                one_count += 1

    assert one_count / total > 0.95, f"Should measure |1⟩ due to interference, got {one_count / total}"


def test_rotation_gates() -> None:
    """Test rotation gates with specific angles."""

    @guppy
    def rotation_circuit() -> bool:
        """Test Y and Z rotations."""
        q = qubit()

        # Rotate around Y axis by π/2 (creates equal superposition)
        # angle takes halfturns, so 0.5 halfturns = π/2
        ry(q, angle(0.5))  # π/2

        # Rotate around Z axis by π/4 (adds phase)
        # 0.25 halfturns = π/4
        rz(q, angle(0.25))  # π/4

        # Measure
        return measure(q)

    # Run simulation with state_vector backend
    results = sim(Guppy(rotation_circuit)).qubits(1).quantum(state_vector()).seed(42).run(1000)

    assert results is not None, "Should get results"

    # After Ry(π/2), should be in equal superposition
    # Rz just adds phase, doesn't change measurement probabilities
    zero_count = 0
    one_count = 0

    if hasattr(results, "__getitem__"):
        measurements = results.get("measurement_0", [])
        for m in measurements:
            if m == 0:
                zero_count += 1
            else:
                one_count += 1

    total = zero_count + one_count
    # Should be roughly 50/50 after Ry(π/2)
    assert 0.4 < zero_count / total < 0.6, f"Should be ~50% |0⟩ after Ry(π/2), got {zero_count / total}"
    assert 0.4 < one_count / total < 0.6, f"Should be ~50% |1⟩ after Ry(π/2), got {one_count / total}"



# Copyright 2018 The PECOS Developers
# Copyright 2018 National Technology & Engineering Solutions of Sandia, LLC (NTESS). Under the terms of Contract
# DE-NA0003525 with NTESS, the U.S. Government retains certain rights in this software.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance with
# the License.You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
# specific language governing permissions and limitations under the License.

"""Check Circuits.

==============

This namespace is for callables that take checks and convert them to physical quantum-circuits.

Callables that can be used for general QECCs should be stored here. If a circuit implementation is specific to the QECC
it may be stored in the QECC's folder.

"""

from __future__ import annotations

from typing import TYPE_CHECKING, Any, TypeVar

from pecos.circuits import QuantumCircuit

if TYPE_CHECKING:
    from pecos.circuits.logical_circuit import LogicalCircuit
    from pecos.protocols import LogicalInstructionProtocol

T = TypeVar("T")


class Check2Circuits:
    """Converts checks to circuits."""

    def __init__(self) -> None:
        """Initialize the Check2Circuits converter.

        Sets the name attribute to "Check2Circuits".
        """
        self.name = "Check2Circuits"

    @staticmethod
    def get_num_ancillas(num_checks: int) -> int:
        """Get the number of ancilla qubits needed for the given number of checks.

        Args:
        ----
            num_checks: The number of checks to be performed.
        """
        return num_checks

    def compile(
        self,
        instr: LogicalInstructionProtocol,
        abstract_circuit: LogicalCircuit,
        mapping: dict[int, int] | None = None,
    ) -> QuantumCircuit:
        """Converts abstract circuits that have checks into an instance of ``QuantumCircuit``.

        Args:
        ----
            instr: The logical instruction being compiled.
            abstract_circuit: Abstract circuit that contains checks.
            mapping: Optional mapping dictionary for qubit indices.

        Returns:
        -------
            ``QuantumCircuit``
        """
        make_ticks = self._check_ticks(abstract_circuit)

        gate_params = instr.gate_params

        if mapping is None:
            mapping = gate_params.get("mapping", NoMap())

        forced_outcome = gate_params.get("forced_outcome", None)

        if forced_outcome is None:
            forced_outcome = abstract_circuit.metadata.get("forced_outcome", True)

        if make_ticks:
            largest_tick = 0

            if make_ticks["max_xdatas"]:
                # init H [data ticks] H meas
                largest_tick = make_ticks["max_xdatas"] + 3

            if make_ticks["max_zdatas"]:
                # init [data ticks] meas
                temp = make_ticks["max_zdatas"] + 1
                largest_tick = max(temp, largest_tick)

            if not make_ticks["max_xdatas"] and not make_ticks["max_zdatas"]:
                msg = "Something very weird happened!"
                raise Exception(msg)

        else:
            largest_tick = 0
            # find maximum tick
            for gate_symbol, _locations, params in abstract_circuit.items():
                if gate_symbol in {"X check", "Z check"}:
                    ancilla_ticks = params["ancilla_ticks"]  # Initialization of the ancilla.
                    data_ticks = params["data_ticks"]
                    meas_ticks = params["meas_ticks"]

                    for ticks in [ancilla_ticks, data_ticks, meas_ticks]:
                        tick_list = ticks
                        if isinstance(tick_list, int):
                            tick_list = [ticks]

                        for t in tick_list:
                            largest_tick = max(t, largest_tick)

                else:
                    tick = params["tick"]

                    largest_tick = max(tick, largest_tick)

        # A quantum circuit with number of ticks == ``largest_tick``
        circuit = QuantumCircuit(largest_tick + 1, **gate_params)

        # Add circuits
        # ============
        for gate_symbol, locations, params in abstract_circuit.items():
            if gate_symbol in {"X check", "Z check"}:
                datas = params["datas"]
                ancillas = params["ancillas"]
                if not isinstance(ancillas, int):
                    if len(ancillas) == 1:
                        ancillas = ancillas[0]
                    else:
                        msg = "This circuit compiler only accepts single ancillas per check!"
                        raise Exception(msg)

                if make_ticks:
                    # Come up with ticks if none have been specified.
                    ancilla_ticks, data_ticks, meas_ticks = self.generate_ticks(
                        make_ticks,
                        gate_symbol,
                        locations,
                        params,
                    )

                else:
                    ancilla_ticks = params["ancilla_ticks"]
                    data_ticks = params["data_ticks"]
                    meas_ticks = params["meas_ticks"]

                # Add ancilla init
                # ----------------
                if isinstance(ancilla_ticks, int):
                    circuit.update(
                        {"init |0>": {mapping[ancillas]}},
                        tick=ancilla_ticks,
                    )

                    if gate_symbol == "X check":
                        circuit.update(
                            {"H": {mapping[ancillas]}},
                            tick=ancilla_ticks + 1,
                        )
                        circuit.update({"H": {mapping[ancillas]}}, tick=meas_ticks - 1)
                else:
                    msg = "Can not currently handle multiple ancilla checks!"
                    raise Exception(msg)

                # Add data
                # --------
                if hasattr(data_ticks, "__iter__"):
                    if gate_symbol == "X check":
                        for i, t in enumerate(data_ticks):
                            circuit.update(
                                {"CNOT": {(mapping[ancillas], mapping[datas[i]])}},
                                tick=t,
                            )
                    elif gate_symbol == "Z check":
                        for i, t in enumerate(data_ticks):
                            circuit.update(
                                {"CNOT": {(mapping[datas[i]], mapping[ancillas])}},
                                tick=t,
                            )
                else:
                    msg = "Can not currently handle single data checks!"
                    raise Exception(msg)

                # Add ancilla measurements
                # ------------------------
                if isinstance(meas_ticks, int):
                    if forced_outcome:
                        circuit.update(
                            {"measure Z": {mapping[ancillas]}},
                            tick=meas_ticks,
                        )
                    else:
                        circuit.update(
                            {"measure Z": {mapping[ancillas]}},
                            tick=meas_ticks,
                            forced_outcome=0,
                        )
                else:
                    msg = "Can not currently handle multiple ancilla checks!"
                    raise Exception(msg)

            else:
                tick = params["tick"]
                circuit.update(
                    {gate_symbol: self.mapset(mapping, set(locations))},
                    tick=tick,
                )

        return circuit  # Return QuantumCircuit and number of ancillas used in this circuit.

    @staticmethod
    def mapset(mapping: dict[int, int], oldset: set[int]) -> set[int]:
        """Applies a mapping to a set.

        Args:
        ----
            mapping: Dictionary mapping old indices to new indices.
            oldset: Set of indices to be mapped.
        """
        newset = set()

        for e in oldset:
            newset.add(mapping[e])

        return newset

    @staticmethod
    def _check_ticks(
        abstract_circuit: LogicalCircuit,
    ) -> tuple[bool, bool, dict[str, Any]]:
        # Determine if any X checks or Z check
        # Determine if ancilla_ticks, data_ticks, or meas_ticks ever set

        has_ticks_set = False
        max_xdatas = 0
        max_zdatas = 0

        for gate_symbol, _, params in abstract_circuit.items():
            if gate_symbol in {"X check", "Z check"}:
                ancilla_ticks = params.get("ancilla_ticks")
                data_ticks = params.get("data_ticks")
                meas_ticks = params.get("meas_ticks")
                datas = params["datas"]

                if ancilla_ticks or data_ticks or meas_ticks:
                    has_ticks_set = True
                    break

                if gate_symbol == "X check":
                    max_xdatas = max(len(datas), max_xdatas)

                elif len(datas) > max_zdatas:
                    max_zdatas = len(datas)

        if has_ticks_set:
            return None

        # ticks init |0>, [H], cnots, [H], meas Z
        return {"max_xdatas": max_xdatas, "max_zdatas": max_zdatas}

    @staticmethod
    def generate_ticks(
        make_ticks_data: dict[str, Any],
        gate_symbol: str,
        _locations: set[int],
        params: dict[str, Any],
    ) -> dict[str, Any]:
        """Generate tick data for stabilizer checks.

        Args:
            make_ticks_data: Data structure containing tick generation parameters.
            gate_symbol: Symbol identifying the type of gate/check to generate.
            locations: Set of qubit locations (unused in this implementation).
            params: Additional parameters for gate generation.

        Returns:
            Dictionary containing generated tick data.
        """
        # X check: init   H    [all data ticks begin] <- H meas [slide to the left]
        # Z check: [idle] init [all data ticks begin] <- meas [slide to the left]

        max_xdatas = make_ticks_data["max_xdatas"]

        if gate_symbol == "X check":
            ancilla_ticks = 0
            data_ticks = [i + 1 for i in range(len(params["datas"]))]
            meas_ticks = data_ticks[-1] + 2  # One extra to include the H
        elif max_xdatas:  # If there are X checks
            ancilla_ticks = 1
            data_ticks = [i + 2 for i in range(len(params["datas"]))]
            meas_ticks = data_ticks[-1] + 1
        else:
            ancilla_ticks = 0
            data_ticks = [i + 1 for i in range(len(params["datas"]))]
            meas_ticks = data_ticks[-1] + 1
        return ancilla_ticks, data_ticks, meas_ticks


class NoMap:
    """Default Mapping: item -> item."""

    def __init__(self) -> None:
        """Initialize the NoMap identity mapping.

        NoMap provides a default identity mapping where each item maps to itself.
        """

    def __getitem__(self, item: T) -> T:
        """Return the item unchanged (identity mapping)."""
        return item



"""Bitflip noise with leakage for qubit initialization.

This module provides noise models for qubit initialization operations that
include both bitflip errors and leakage to states outside the computational
subspace, providing more realistic error modeling.
"""

# Copyright 2023 The PECOS Developers
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance with
# the License.You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
# specific language governing permissions and limitations under the License.

from __future__ import annotations

from typing import TYPE_CHECKING

from pecos.error_models.noise_impl.noise_initz_bitflip import noise_initz_bitflip
from pecos.reps.pyphir.op_types import QOp

if TYPE_CHECKING:
    from pecos.protocols import MachineProtocol


def noise_initz_bitflip_leakage(op: QOp, p: float, machine: MachineProtocol) -> None:
    """The noise model for qubit (re)initialization, including leakage support.

    Args:
    ----
        op: Ideal quantum operation.
        p: Probability of bitflip.
        machine: Machine protocol instance containing leakage state information.
    """
    args = set(op.args)
    leaked = machine.leaked_qubits & args

    noise = []
    noisy_ops = machine.unleak(leaked)
    if noise:
        noise.extend(noisy_ops)

    not_leaked = args - leaked
    if not_leaked:
        remaining_inits = QOp(
            name=op.name,
            args=list(not_leaked),
            metadata=dict(op.metadata),
        )
        noisy_ops = noise_initz_bitflip(remaining_inits, p)
        if noisy_ops:
            noise.extend(noisy_ops)

    return noise



# Copyright 2022 The PECOS Developers
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance with
# the License.You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
# specific language governing permissions and limitations under the License.

"""Classical computation utilities for the PECOS virtual machine.

This module provides functions for evaluating classical operations, expressions,
and conditional logic in the PECOS classical virtual machine (CVM).
"""

from __future__ import annotations

from typing import TYPE_CHECKING

from pecos import BitInt

if TYPE_CHECKING:
    from typing import Any

    from pecos.circuits import QuantumCircuit
    from pecos.protocols import SimulatorProtocol


def set_output(
    state: SimulatorProtocol,
    circuit: QuantumCircuit,
    output_spec: dict[str, int] | None,
    output: dict[str, BitInt] | None,
) -> dict[str, BitInt]:
    """Set up output dictionary for classical variable storage.

    Initializes the output dictionary with BitInts for storing classical
    computation results, using size specifications from the circuit metadata
    and provided output specification.

    Args:
        state: Quantum simulator state providing qubit count.
        circuit: Quantum circuit containing variable specifications.
        output_spec: Dictionary mapping variable names to bit sizes.
        output: Existing output dictionary to update, if any.

    Returns:
        Initialized output dictionary with BitInts for each variable.
    """
    if output_spec is None:
        output_spec = {}

    output_spec["__pecos_scratch"] = state.num_qubits

    if circuit.metadata.get("cvar_spec"):
        output_spec_new = circuit.metadata["cvar_spec"]
        output_spec_new.update(output_spec)
        output_spec = output_spec_new

    if output is None:
        output = {}

        if output_spec:
            for symbol, size in output_spec.items():
                output[symbol] = BitInt(size)

    return output


def eval_op(
    op: str,
    a: BitInt | int,
    b: BitInt | int | None = None,
    width: int = 32,
) -> BitInt:
    """Evaluate a binary or unary operation on BitInts.

    Performs arithmetic, logical, or comparison operations on binary arrays,
    supporting assignment, bitwise operations, arithmetic, and comparisons.

    Args:
        op: Operation string (e.g., '=', '+', '&', '==', '~').
        a: First operand as BitInt or integer.
        b: Second operand for binary operations, None for unary operations.
        width: Bit width for integer to BitInt conversion.

    Returns:
        Result of the operation as a BitInt.

    Raises:
        Exception: If operation is unsupported or arguments are invalid.
    """
    if isinstance(a, int):
        a = BitInt(width, a)

    if op == "=":
        if b:
            msg = "Assignment can only have one argument (only `a`)."
            raise Exception(msg)

        return a

    if op == "|":
        expr_eval = a | b
    elif op == "^":
        expr_eval = a ^ b
    elif op == "&":
        expr_eval = a & b
    elif op == "+":
        expr_eval = a + b
    elif op == "-":
        expr_eval = a - b
    elif op == ">>":
        expr_eval = a >> b
    elif op == "<<":
        expr_eval = a << b
    elif op == "*":
        expr_eval = a * b
    elif op == "/":
        expr_eval = a // b
    elif op == "==":
        expr_eval = a == b
    elif op == "!=":
        expr_eval = a != b
    elif op == "<=":
        expr_eval = a <= b
    elif op == ">=":
        expr_eval = a >= b
    elif op == "<":
        expr_eval = a < b
    elif op == ">":
        expr_eval = a > b
    elif op == "%":
        expr_eval = a % b

    elif op == "~":
        expr_eval = ~a

        if b:
            msg = "Unary operation but got another argument!!!."
            raise Exception(msg)

    else:
        msg = (
            f'Receive op "{op}". Only operators `=`, `~`, `|`, `^`, `&`, `+`, `-`, `<<`, and `>>` '
            f"have been implemented."
        )
        raise Exception(msg)

    return expr_eval


def get_val(
    a: BitInt | tuple[str, int] | list[str | int] | str | int,
    output: dict[str, BitInt],
    width: int,
    shot_id: int,
) -> BitInt:
    """Extract and convert a value to BitInt.

    Retrieves values from the output dictionary or converts literals to BitInts,
    supporting indexed access for array variables.

    Args:
        a: Value to extract - can be BitInt, variable reference, or literal.
        output: Dictionary containing variable values.
        width: Bit width for value conversion.
        shot_id: The current instance's shot id

    Returns:
        Value converted to BitInt format.

    Raises:
        TypeError: If the input type is not supported.
    """
    if isinstance(a, BitInt):
        return a

    if isinstance(a, tuple | list):
        sym, idx = a
        val = output[sym][idx]

    elif isinstance(a, str):
        val = shot_id if a == "JOB_shotnum" else int(output[a])

    elif isinstance(a, int):
        val = a

    else:
        msg = f'Could not evaluate "{a!s}". Wrong type, got type: {type(a)}.'
        raise TypeError(msg)

    return BitInt(width, val)


def recur_eval_op(
    expr_dict: dict[str, Any],
    output: dict[str, BitInt],
    width: int,
    shot_id: int,
) -> BitInt:
    """Recursively evaluate a nested expression dictionary.

    Processes nested expressions by recursively evaluating sub-expressions
    and combining results using the specified operations.

    Args:
        expr_dict: Dictionary containing expression with 'op', 'a', 'b', 'c' keys.
        output: Dictionary containing variable values.
        width: Bit width for operations.
        shot_id: The current instance's shot id.

    Returns:
        Result of the evaluated expression as BitInt.
    """
    a = expr_dict.get("a")
    op = expr_dict.get("op")
    b = expr_dict.get("b")
    c = expr_dict.get("c")

    if isinstance(a, dict):
        a = recur_eval_op(a, output, width, shot_id=shot_id)

    elif c:  # c => unary operation
        c = (
            recur_eval_op(c, output, width, shot_id=shot_id)
            if isinstance(c, dict)
            else get_val(c, output, width, shot_id)
        )

        a = eval_op(op, c, width=width)

    else:
        a = get_val(a, output, width, shot_id)

    if b:
        b = (
            recur_eval_op(b, output, width, shot_id=shot_id)
            if isinstance(b, dict)
            else get_val(b, output, width, shot_id)
        )

        a = eval_op(op, a, b, width=width)

    return a


def eval_cop(
    cop_expr: dict[str, Any] | list[dict[str, Any]],
    output: dict[str, BitInt],
    width: int,
    shot_id: int,
) -> None:
    """Evaluate classical operation expression.

    Evaluate classical expression such as:

    assignment:
    t = a     BitInt = (BitInt | int)
    t[i] = a  BitInt[i] = (BitInt | int)

    binary operations:
    t = a o b
    t[i] = a[j] o b[k]
    """
    # Get `t` argument
    # ----------------
    t = cop_expr["t"]  # symbol of where the resulting value will be stored in the output

    if isinstance(t, str):
        t_sym = t
        t_index = None
    elif isinstance(t, tuple | list) and len(t) == 2:
        t_sym = t[0]
        t_index = t[1]
    else:
        msg = "`t` should be `str` or `Tuple[str, int]`!"
        raise Exception(msg)

    t_obj = output[t_sym]

    # Eval assignment
    # ---------------
    expr_eval = recur_eval_op(cop_expr, output, width, shot_id=shot_id)

    # Assign the final value:
    # -----------------------
    if t_index is not None:  # t[i] = ...
        t_obj[t_index] = expr_eval[0]

    else:  # t = ...
        t_obj.set_clip(expr_eval)


def eval_tick_conds(
    tick_circuit: QuantumCircuit,
    output: dict[str, BitInt],
) -> list[bool]:
    """Evaluate conditional expressions for each operation in a tick circuit.

    Processes each operation in the circuit and evaluates its conditional
    expression to determine if the operation should be executed.

    Args:
        tick_circuit: Quantum circuit containing operations with conditions.
        output: Dictionary containing variable values for condition evaluation.

    Returns:
        List of boolean values indicating whether each operation's condition is true.
    """
    conds = []

    for _symbol, _locations, params in tick_circuit.items():
        cond_eval = eval_condition(params.get("cond"), output)

        conds.append(cond_eval)
    return conds


def eval_condition(
    conditional_expr: dict[str, Any] | tuple[Any, ...] | list[Any] | None,
    output: dict[str, BitInt],
) -> bool:
    """Evaluate a conditional expression to a boolean result.

    Processes conditional expressions supporting comparison operators,
    variable references, and nested conditions. Returns True for None conditions.

    Args:
        conditional_expr: Expression dictionary with 'op', 'a', 'b' keys,
                         tuple/list for complex conditions, or None.
        output: Dictionary containing variable values.

    Returns:
        Boolean result of the conditional evaluation.

    Raises:
        Exception: If expression format is invalid.
        TypeError: If operand types are unexpected.
    """
    # Handle if a condition might eval to something else (eval_to)
    if isinstance(conditional_expr, tuple | list):
        if len(conditional_expr) != 2:
            msg = "Not expected conditional to have more than 2 elements."
            raise Exception(msg)

        if not isinstance(conditional_expr[1], bool):
            msg = "Expecting the second conditional element to be bool."
            raise TypeError(msg)

        return eval_condition(conditional_expr[0], output) == eval_condition(
            conditional_expr[1],
            output,
        )

    if conditional_expr:
        a = conditional_expr["a"]
        b = conditional_expr["b"]
        op = conditional_expr["op"]
        if isinstance(a, str):
            a = output[a]  # str -> BitInt
        elif isinstance(a, tuple | list) and len(a) == 2:
            a = output[a[0]][a[1]]  # (str, int) -> int (1 or 0)
        else:
            msg = "`a` should be `str` or `Tuple[str, int]`!"
            raise Exception(msg)

        if isinstance(b, str):
            b = output[b]  # str -> BitInt
        elif isinstance(b, tuple | list) and len(b) == 2:
            b = output[b[0]][b[1]]  # (str, int) -> int (1 or 0)
        elif isinstance(b, int):
            pass
        else:
            msg = "`b` should be `str` or `Tuple[str, int]` or `int`!"
            raise Exception(msg)

        # Map of operators to their evaluation functions
        ops = {
            "==": lambda a, b: bool(a == b),
            "!=": lambda a, b: bool(a != b),
            "^": lambda a, b: bool(int(a ^ b)),
            "|": lambda a, b: bool(int(a | b)),
            "&": lambda a, b: bool(int(a & b)),
            "<": lambda a, b: a < b,
            ">": lambda a, b: a > b,
            "<=": lambda a, b: a <= b,
            ">=": lambda a, b: a >= b,
            ">>": lambda a, b: a >> b,
            "<<": lambda a, b: a << b,
            "~": lambda a, _: ~a,
            "*": lambda a, b: a * b,
            "/": lambda a, b: a // b,
        }

        if op in ops:
            return ops[op](a, b)

        msg = "Comparison operator not recognized!"
        raise Exception(msg)

    return True



# Copyright 2022 The PECOS Developers
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance with
# the License.You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
# specific language governing permissions and limitations under the License.

"""Classical Virtual Machine (CVM) implementation for PECOS.

This module provides the main CVM implementation for executing classical
computations alongside quantum operations in hybrid quantum-classical algorithms.
"""

from __future__ import annotations

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from collections.abc import Callable
    from typing import Any, Protocol

    class CCOPVMProtocol(Protocol):
        """Protocol for CCOP VM objects."""

        def exec(self, func_name: str, args: list) -> int | None:
            """Execute a function in the classical co-processor VM.

            Args:
                func_name: Name of the function to execute.
                args: List of arguments to pass to the function.

            Returns:
                Optional integer result from the function execution.
            """
            ...

    class InterpreterProtocol(Protocol):
        """Protocol for interpreter objects."""

        # Without more context, we'll keep this minimal


class CVM:
    """Classical Virtual Machine for executing classical functions and statements.

    The CVM manages the execution of classical code within the hybrid quantum-classical
    computing framework, providing interfaces for classical operations and interpreters.
    """

    def __init__(
        self,
        ccop_vm: CCOPVMProtocol | None = None,
        cinterpreter: InterpreterProtocol | None = None,
        sim_debug: dict[str, Callable[..., Any]] | None = None,
    ) -> None:
        """Classical Virtual Machine, which is responsible for executing classical functions and statements.

        Attributes:
        ----------
            ccop_vm: A VM representing the computing environment of a classical co-processor. This generally provides
                external classical functions that are usually problem specific.
            cinterpreter: Provides an interpreter for generic classical statements. For example, boolean operations,
                assignments, comparisons, etc.
            sim_debug: A collection of functions used in the simulation environment to provide additional information
                that may not typically be available to a physical quantum device.
        """
        self.state = {}

        self.ccop_vm = ccop_vm
        self.cinterpreter = cinterpreter
        self.sim_debug = sim_debug

    def reset_state(self) -> None:
        """Reset the CVM state to its initial empty state."""
        self.state = ()

    def exec(self, func_name: str, args: list[Any]) -> None:
        """Execute a function in the classical virtual machine.

        Args:
            func_name: Name of the function to execute.
            args: List of arguments to pass to the function.
        """



"""Bitflip noise with leakage for measurement operations.

This module provides noise models for quantum measurement operations that
include both bitflip errors and leakage effects, providing comprehensive
error modeling for measurement processes.
"""

# Copyright 2023 The PECOS Developers
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance with
# the License.You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
# specific language governing permissions and limitations under the License.

from __future__ import annotations

from typing import TYPE_CHECKING

import pecos as pc
from pecos.reps.pyphir.op_types import QOp

if TYPE_CHECKING:
    from pecos.protocols import MachineProtocol


def noise_meas_bitflip_leakage(
    op: QOp,
    p: float,
    machine: MachineProtocol,
) -> list[QOp] | None:
    """Bit-flip noise model for measurements.

    Args:
    ----
        op: Ideal quantum operation.
        p: measurement error rate.
        machine: Machine protocol instance containing leakage state information.
    """
    # Bit flip noise
    # --------------
    # Use fused operation to check and get error indices in one pass
    error_indices = pc.random.compare_indices(len(op.args), p)

    noise = []

    leakded = machine.leaked_qubits & set(op.args)
    if leakded:
        noisy_ops = machine.meas_leaked(leakded)
        noise.extend(noisy_ops)

    if error_indices:
        bitflips = [op.args[idx] for idx in error_indices]

        noisy_op = QOp(
            name="Measure",
            args=list(op.args),
            returns=list(op.returns),
            metadata=dict(op.metadata),
        )
        noisy_op.metadata["bitflips"] = bitflips
        noise.append(noisy_op)

        return noise

    if noise:
        return noise
    return None



"""Comprehensive gate verification tests for the HUGR interpreter.

Tests that all Guppy quantum gate types execute correctly and produce
the expected quantum state transformations.
"""

import pytest
from guppylang import guppy
from guppylang.std.quantum import (
    crz,
    cx,
    cy,
    cz,
    discard,
    h,
    measure,
    pi,
    qubit,
    reset,
    rx,
    ry,
    rz,
    s,
    sdg,
    t,
    tdg,
    toffoli,
    v,
    vdg,
    x,
    y,
    z,
)
from pecos import Guppy, sim
from pecos_rslib import state_vector


def run_circuit(
    guppy_func: object,
    num_qubits: int,
    shots: int = 100,
    seed: int = 42,
) -> dict:
    """Run a Guppy function with PECOS direct HUGR interpreter."""
    return sim(Guppy(guppy_func)).qubits(num_qubits).quantum(state_vector()).seed(seed).run(shots).to_dict()


def get_measurements(results: dict) -> list:
    """Extract measurements from results."""
    return results.get("measurements", [])


class TestSingleQubitGates:
    """Test single-qubit gates produce correct state transformations."""

    def test_x_gate(self) -> None:
        """X gate: |0> -> |1>."""

        @guppy
        def circuit() -> bool:
            q = qubit()
            x(q)
            return measure(q)

        measurements = get_measurements(run_circuit(circuit, 1))
        assert all(m == [1] for m in measurements)

    def test_y_gate(self) -> None:
        """Y gate: |0> -> i|1> (measures as |1>)."""

        @guppy
        def circuit() -> bool:
            q = qubit()
            y(q)
            return measure(q)

        measurements = get_measurements(run_circuit(circuit, 1))
        assert all(m == [1] for m in measurements)

    def test_z_gate_on_zero(self) -> None:
        """Z gate: |0> -> |0> (no change in measurement)."""

        @guppy
        def circuit() -> bool:
            q = qubit()
            z(q)
            return measure(q)

        measurements = get_measurements(run_circuit(circuit, 1))
        assert all(m == [0] for m in measurements)

    def test_z_gate_on_one(self) -> None:
        """Z gate: |1> -> -|1> (still measures as |1>)."""

        @guppy
        def circuit() -> bool:
            q = qubit()
            x(q)
            z(q)
            return measure(q)

        measurements = get_measurements(run_circuit(circuit, 1))
        assert all(m == [1] for m in measurements)

    def test_h_gate_superposition(self) -> None:
        """H gate: |0> -> |+> (50/50 distribution)."""

        @guppy
        def circuit() -> bool:
            q = qubit()
            h(q)
            return measure(q)

        measurements = get_measurements(run_circuit(circuit, 1, shots=1000))
        zeros = sum(1 for m in measurements if m == [0])
        ones = sum(1 for m in measurements if m == [1])
        # Expect roughly 50/50 with some tolerance
        assert 400 < zeros < 600, f"Expected ~500 zeros, got {zeros}"
        assert 400 < ones < 600, f"Expected ~500 ones, got {ones}"

    def test_h_h_identity(self) -> None:
        """H-H = I: |0> -> |+> -> |0>."""

        @guppy
        def circuit() -> bool:
            q = qubit()
            h(q)
            h(q)
            return measure(q)

        measurements = get_measurements(run_circuit(circuit, 1))
        assert all(m == [0] for m in measurements)

    def test_s_gate(self) -> None:
        """S gate (phase): |0> -> |0>."""

        @guppy
        def circuit() -> bool:
            q = qubit()
            s(q)
            return measure(q)

        measurements = get_measurements(run_circuit(circuit, 1))
        assert all(m == [0] for m in measurements)

    def test_sdg_gate(self) -> None:
        """Sdg gate (S-dagger): |0> -> |0>."""

        @guppy
        def circuit() -> bool:
            q = qubit()
            sdg(q)
            return measure(q)

        measurements = get_measurements(run_circuit(circuit, 1))
        assert all(m == [0] for m in measurements)

    def test_s_sdg_identity(self) -> None:
        """S-Sdg = I on superposition."""

        @guppy
        def circuit() -> bool:
            q = qubit()
            h(q)
            s(q)
            sdg(q)
            h(q)
            return measure(q)

        measurements = get_measurements(run_circuit(circuit, 1))
        assert all(m == [0] for m in measurements)

    def test_t_gate(self) -> None:
        """T gate: |0> -> |0>."""

        @guppy
        def circuit() -> bool:
            q = qubit()
            t(q)
            return measure(q)

        measurements = get_measurements(run_circuit(circuit, 1))
        assert all(m == [0] for m in measurements)

    def test_tdg_gate(self) -> None:
        """Tdg gate (T-dagger): |0> -> |0>."""

        @guppy
        def circuit() -> bool:
            q = qubit()
            tdg(q)
            return measure(q)

        measurements = get_measurements(run_circuit(circuit, 1))
        assert all(m == [0] for m in measurements)

    def test_t_tdg_identity(self) -> None:
        """T-Tdg = I on superposition."""

        @guppy
        def circuit() -> bool:
            q = qubit()
            h(q)
            t(q)
            tdg(q)
            h(q)
            return measure(q)

        measurements = get_measurements(run_circuit(circuit, 1))
        assert all(m == [0] for m in measurements)

    def test_v_gate_squared(self) -> None:
        """V gate (sqrt(X)): V^2 = X, so |0> -> |1>."""

        @guppy
        def circuit() -> bool:
            q = qubit()
            v(q)
            v(q)
            return measure(q)

        measurements = get_measurements(run_circuit(circuit, 1))
        assert all(m == [1] for m in measurements)

    def test_vdg_gate_squared(self) -> None:
        """Vdg gate: Vdg^2 = X, so |0> -> |1>."""

        @guppy
        def circuit() -> bool:
            q = qubit()
            vdg(q)
            vdg(q)
            return measure(q)

        measurements = get_measurements(run_circuit(circuit, 1))
        assert all(m == [1] for m in measurements)

    def test_v_vdg_identity(self) -> None:
        """V-Vdg = I."""

        @guppy
        def circuit() -> bool:
            q = qubit()
            v(q)
            vdg(q)
            return measure(q)

        measurements = get_measurements(run_circuit(circuit, 1))
        assert all(m == [0] for m in measurements)


class TestRotationGates:
    """Test rotation gates with angle parameters."""

    def test_rx_pi(self) -> None:
        """RX(pi) = X (up to global phase)."""

        @guppy
        def circuit() -> bool:
            q = qubit()
            rx(q, pi)
            return measure(q)

        measurements = get_measurements(run_circuit(circuit, 1))
        assert all(m == [1] for m in measurements)

    def test_ry_pi(self) -> None:
        """RY(pi) = Y (up to global phase)."""

        @guppy
        def circuit() -> bool:
            q = qubit()
            ry(q, pi)
            return measure(q)

        measurements = get_measurements(run_circuit(circuit, 1))
        assert all(m == [1] for m in measurements)

    def test_rz_on_zero(self) -> None:
        """RZ(pi) on |0> = |0>."""

        @guppy
        def circuit() -> bool:
            q = qubit()
            rz(q, pi)
            return measure(q)

        measurements = get_measurements(run_circuit(circuit, 1))
        assert all(m == [0] for m in measurements)

    def test_rz_phase_flip(self) -> None:
        """RZ(pi): |+> -> |-> -> |1> via H-RZ-H."""

        @guppy
        def circuit() -> bool:
            q = qubit()
            h(q)
            rz(q, pi)
            h(q)
            return measure(q)

        measurements = get_measurements(run_circuit(circuit, 1))
        assert all(m == [1] for m in measurements)


class TestTwoQubitGates:
    """Test two-qubit gates."""

    def test_cx_control_zero(self) -> None:
        """CX: control=|0>, no flip."""

        @guppy
        def circuit() -> tuple[bool, bool]:
            q0 = qubit()
            q1 = qubit()
            cx(q0, q1)
            return measure(q0), measure(q1)

        measurements = get_measurements(run_circuit(circuit, 2))
        assert all(m == [0, 0] for m in measurements)

    def test_cx_control_one(self) -> None:
        """CX: control=|1>, flip target."""

        @guppy
        def circuit() -> tuple[bool, bool]:
            q0 = qubit()
            q1 = qubit()
            x(q0)
            cx(q0, q1)
            return measure(q0), measure(q1)

        measurements = get_measurements(run_circuit(circuit, 2))
        assert all(m == [1, 1] for m in measurements)

    def test_cy_control_zero(self) -> None:
        """CY: control=|0>, no flip."""

        @guppy
        def circuit() -> tuple[bool, bool]:
            q0 = qubit()
            q1 = qubit()
            cy(q0, q1)
            return measure(q0), measure(q1)

        measurements = get_measurements(run_circuit(circuit, 2))
        assert all(m == [0, 0] for m in measurements)

    def test_cy_control_one(self) -> None:
        """CY: control=|1>, flip target."""

        @guppy
        def circuit() -> tuple[bool, bool]:
            q0 = qubit()
            q1 = qubit()
            x(q0)
            cy(q0, q1)
            return measure(q0), measure(q1)

        measurements = get_measurements(run_circuit(circuit, 2))
        assert all(m == [1, 1] for m in measurements)

    def test_cz_phase_only(self) -> None:
        """CZ: both |1>, phase only (no measurement change)."""

        @guppy
        def circuit() -> tuple[bool, bool]:
            q0 = qubit()
            q1 = qubit()
            x(q0)
            x(q1)
            cz(q0, q1)
            return measure(q0), measure(q1)

        measurements = get_measurements(run_circuit(circuit, 2))
        assert all(m == [1, 1] for m in measurements)

    def test_crz_control_zero(self) -> None:
        """CRZ: control=|0>, no effect."""

        @guppy
        def circuit() -> tuple[bool, bool]:
            q0 = qubit()
            q1 = qubit()
            crz(q0, q1, pi)
            return measure(q0), measure(q1)

        measurements = get_measurements(run_circuit(circuit, 2))
        assert all(m == [0, 0] for m in measurements)

    def test_crz_control_one(self) -> None:
        """CRZ: control=|1>, applies RZ(pi) to target."""

        @guppy
        def circuit() -> tuple[bool, bool]:
            q0 = qubit()
            q1 = qubit()
            x(q0)  # ctrl = |1>
            h(q1)  # target = |+>
            crz(q0, q1, pi)  # target -> |->
            h(q1)  # target -> |1>
            return measure(q0), measure(q1)

        measurements = get_measurements(run_circuit(circuit, 2))
        assert all(m == [1, 1] for m in measurements)


class TestThreeQubitGates:
    """Test three-qubit gates (Toffoli)."""

    def test_toffoli_00(self) -> None:
        """Toffoli: controls=00, no flip."""

        @guppy
        def circuit() -> tuple[bool, bool, bool]:
            q0 = qubit()
            q1 = qubit()
            q2 = qubit()
            toffoli(q0, q1, q2)
            return measure(q0), measure(q1), measure(q2)

        measurements = get_measurements(run_circuit(circuit, 3))
        assert all(m == [0, 0, 0] for m in measurements)

    def test_toffoli_10(self) -> None:
        """Toffoli: controls=10, no flip."""

        @guppy
        def circuit() -> tuple[bool, bool, bool]:
            q0 = qubit()
            q1 = qubit()
            q2 = qubit()
            x(q0)
            toffoli(q0, q1, q2)
            return measure(q0), measure(q1), measure(q2)

        measurements = get_measurements(run_circuit(circuit, 3))
        assert all(m == [1, 0, 0] for m in measurements)

    def test_toffoli_01(self) -> None:
        """Toffoli: controls=01, no flip."""

        @guppy
        def circuit() -> tuple[bool, bool, bool]:
            q0 = qubit()
            q1 = qubit()
            q2 = qubit()
            x(q1)
            toffoli(q0, q1, q2)
            return measure(q0), measure(q1), measure(q2)

        measurements = get_measurements(run_circuit(circuit, 3))
        assert all(m == [0, 1, 0] for m in measurements)

    def test_toffoli_11(self) -> None:
        """Toffoli: controls=11, flip target."""

        @guppy
        def circuit() -> tuple[bool, bool, bool]:
            q0 = qubit()
            q1 = qubit()
            q2 = qubit()
            x(q0)
            x(q1)
            toffoli(q0, q1, q2)
            return measure(q0), measure(q1), measure(q2)

        measurements = get_measurements(run_circuit(circuit, 3))
        assert all(m == [1, 1, 1] for m in measurements)


class TestResetAndDiscard:
    """Test reset and discard operations."""

    def test_reset_from_one(self) -> None:
        """Reset: |1> -> |0>."""

        @guppy
        def circuit() -> bool:
            q = qubit()
            x(q)
            reset(q)
            return measure(q)

        measurements = get_measurements(run_circuit(circuit, 1))
        assert all(m == [0] for m in measurements)

    def test_reset_from_zero(self) -> None:
        """Reset: |0> -> |0>."""

        @guppy
        def circuit() -> bool:
            q = qubit()
            reset(q)
            return measure(q)

        measurements = get_measurements(run_circuit(circuit, 1))
        assert all(m == [0] for m in measurements)

    def test_discard(self) -> None:
        """Discard doesn't crash and other qubits work."""

        @guppy
        def circuit() -> bool:
            q1 = qubit()
            q2 = qubit()
            x(q1)
            discard(q2)
            return measure(q1)

        measurements = get_measurements(run_circuit(circuit, 2))
        assert all(m == [1] for m in measurements)


class TestEntanglement:
    """Test entangled states."""

    def test_bell_state(self) -> None:
        """Bell state: perfectly correlated 00 or 11."""

        @guppy
        def circuit() -> tuple[bool, bool]:
            q0 = qubit()
            q1 = qubit()
            h(q0)
            cx(q0, q1)
            return measure(q0), measure(q1)

        measurements = get_measurements(run_circuit(circuit, 2, shots=1000))
        # All measurements should be correlated
        assert all(m[0] == m[1] for m in measurements)
        # Should see both 00 and 11
        has_zeros = any(m[0] == 0 for m in measurements)
        has_ones = any(m[0] == 1 for m in measurements)
        assert has_zeros
        assert has_ones

    def test_ghz_3_state(self) -> None:
        """3-qubit GHZ state: perfectly correlated 000 or 111."""

        @guppy
        def circuit() -> tuple[bool, bool, bool]:
            q0 = qubit()
            q1 = qubit()
            q2 = qubit()
            h(q0)
            cx(q0, q1)
            cx(q1, q2)
            return measure(q0), measure(q1), measure(q2)

        measurements = get_measurements(run_circuit(circuit, 3, shots=1000))
        # All measurements should be correlated
        assert all(m[0] == m[1] == m[2] for m in measurements)
        # Should see both 000 and 111
        has_zeros = any(m[0] == 0 for m in measurements)
        has_ones = any(m[0] == 1 for m in measurements)
        assert has_zeros
        assert has_ones


if __name__ == "__main__":
    pytest.main([__file__, "-v"])



"""Bare syndrome extraction gates for surface codes."""

from pecos.slr.qeclib.surface.patches.patch_base import SurfacePatch


class SurfaceBareSynGates:
    """Collection of bare syndrome extraction gates for surface code patches."""

    @staticmethod
    def syn_extr(*patches: SurfacePatch, rounds: int = 1) -> None:
        """Measure `rounds` number of syndrome extraction of X and Z checks using bare ancillas."""
        # TODO: ...

    @staticmethod
    def qec(*patches: SurfacePatch) -> list[None]:
        """Run distance number of rounds of syndrome extraction."""
        return [SurfaceBareSynGates.syn_extr(rounds=p.distance) for p in patches]



# Copyright 2023 The PECOS Developers
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance with
# the License.You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
# specific language governing permissions and limitations under the License.

"""Two-qubit gate operations for cuStateVec simulator.

This module provides GPU-accelerated two-qubit quantum gate operations for the NVIDIA cuStateVec simulator,
including CNOT gates, controlled gates, and other entangling operations using CUDA acceleration.
"""

from __future__ import annotations

from typing import TYPE_CHECKING

import cupy as cp
import numpy as np

if TYPE_CHECKING:
    from pecos.simulators.custatevec.state import CuStateVec
    from pecos.typing import SimulatorGateParams
from cuquantum.bindings import custatevec as cusv

import pecos as pc
from pecos.simulators.custatevec.gates_one_qubit import H


def _apply_controlled_matrix(
    state: CuStateVec,
    control: int,
    target: int,
    matrix: cp.ndarray,
) -> None:
    """Apply the matrix to the state. This should be faster for controlled gates.

    Args:
        state: An instance of CuStateVec
        control: The index of the qubit that acts as the control
        target: The index of the qubit that acts as the target
        matrix: The matrix to be applied
    """
    if control >= state.num_qubits or control < 0:
        msg = f"Qubit {control} out of range."
        raise ValueError(msg)
    if target >= state.num_qubits or target < 0:
        msg = f"Qubit {target} out of range."
        raise ValueError(msg)
    # CuStateVec uses smaller qubit index as least significant
    control = state.num_qubits - 1 - control
    target = state.num_qubits - 1 - target

    cusv.apply_matrix(
        handle=state.libhandle,
        sv=state.cupy_vector.data.ptr,
        sv_data_type=state.cuda_type,
        n_index_bits=state.num_qubits,
        matrix=matrix.data.ptr,
        matrix_data_type=state.cuda_type,
        layout=cusv.MatrixLayout.ROW,
        adjoint=0,  # Don't use the adjoint
        targets=[target],
        n_targets=1,
        controls=[control],
        control_bit_values=[],  # No value of control bit assigned
        n_controls=1,
        compute_type=state.compute_type,
        extra_workspace=0,  # Let cuQuantum use the mempool we configured
        extra_workspace_size_in_bytes=0,  # Let cuQuantum use the mempool we configured
    )
    state.stream.synchronize()


def CX(
    state: CuStateVec,
    qubits: tuple[int, int],
    **_params: SimulatorGateParams,
) -> None:
    """Apply controlled X gate.

    Args:
        state: An instance of CuStateVec
        qubits: A tuple with the index of the qubits where the gate is applied
            The one at `qubits[0]` is the control qubit.
    """
    matrix = cp.asarray(
        [
            0,
            1,
            1,
            0,
        ],
        dtype=state.cp_type,
    )
    _apply_controlled_matrix(state, qubits[0], qubits[1], matrix)


def CY(
    state: CuStateVec,
    qubits: tuple[int, int],
    **_params: SimulatorGateParams,
) -> None:
    """Apply controlled Y gate.

    Args:
        state: An instance of CuStateVec
        qubits: A tuple with the index of the qubits where the gate is applied
            The one at `qubits[0]` is the control qubit.
    """
    matrix = cp.asarray(
        [
            0,
            -1j,
            1j,
            0,
        ],
        dtype=state.cp_type,
    )
    _apply_controlled_matrix(state, qubits[0], qubits[1], matrix)


def CZ(
    state: CuStateVec,
    qubits: tuple[int, int],
    **_params: SimulatorGateParams,
) -> None:
    """Apply controlled Z gate.

    Args:
        state: An instance of CuStateVec
        qubits: A tuple with the index of the qubits where the gate is applied
            The one at `qubits[0]` is the control qubit.
    """
    matrix = cp.asarray(
        [
            1,
            0,
            0,
            -1,
        ],
        dtype=state.cp_type,
    )
    _apply_controlled_matrix(state, qubits[0], qubits[1], matrix)


def _apply_two_qubit_matrix(
    state: CuStateVec,
    qubits: tuple[int, int],
    matrix: cp.ndarray,
) -> None:
    """Apply the matrix to the state.

    Args:
        state: An instance of CuStateVec
        qubits: A tuple of two qubit indices where the gate is applied
        matrix: The matrix to be applied
    """
    if qubits[0] >= state.num_qubits or qubits[0] < 0:
        msg = f"Qubit {qubits[0]} out of range."
        raise ValueError(msg)
    if qubits[1] >= state.num_qubits or qubits[1] < 0:
        msg = f"Qubit {qubits[1]} out of range."
        raise ValueError(msg)
    # CuStateVec uses smaller qubit index as least significant
    q0 = state.num_qubits - 1 - qubits[0]
    q1 = state.num_qubits - 1 - qubits[1]

    cusv.apply_matrix(
        handle=state.libhandle,
        sv=state.cupy_vector.data.ptr,
        sv_data_type=state.cuda_type,
        n_index_bits=state.num_qubits,
        matrix=matrix.data.ptr,
        matrix_data_type=state.cuda_type,
        layout=cusv.MatrixLayout.ROW,
        adjoint=0,  # Don't use the adjoint
        targets=[q0, q1],
        n_targets=2,
        controls=[],
        control_bit_values=[],  # No value of control bit assigned
        n_controls=0,
        compute_type=state.compute_type,
        extra_workspace=0,  # Let cuQuantum use the mempool we configured
        extra_workspace_size_in_bytes=0,  # Let cuQuantum use the mempool we configured
    )
    state.stream.synchronize()


def RXX(
    state: CuStateVec,
    qubits: tuple[int, int],
    angles: tuple[float],
    **_params: SimulatorGateParams,
) -> None:
    """Apply a rotation about XX.

    Args:
        state: An instance of CuStateVec
        qubits: A tuple with the index of the qubits where the gate is applied
        angles: A tuple containing a single angle in radians
    """
    if len(angles) != 1:
        msg = "Gate must be given 1 angle parameter."
        raise ValueError(msg)
    theta = angles[0]

    matrix = cp.asarray(
        [
            pc.cos(theta / 2),
            0,
            0,
            -1j * pc.sin(theta / 2),
            0,
            pc.cos(theta / 2),
            -1j * pc.sin(theta / 2),
            0,
            0,
            -1j * pc.sin(theta / 2),
            pc.cos(theta / 2),
            0,
            -1j * pc.sin(theta / 2),
            0,
            0,
            pc.cos(theta / 2),
        ],
        dtype=state.cp_type,
    )
    _apply_two_qubit_matrix(state, qubits, matrix)


def RYY(
    state: CuStateVec,
    qubits: tuple[int, int],
    angles: tuple[float],
    **_params: SimulatorGateParams,
) -> None:
    """Apply a rotation about YY.

    Args:
        state: An instance of CuStateVec
        qubits: A tuple with the index of the qubits where the gate is applied
        angles: A tuple containing a single angle in radians
    """
    if len(angles) != 1:
        msg = "Gate must be given 1 angle parameter."
        raise ValueError(msg)
    theta = angles[0]

    matrix = cp.asarray(
        [
            pc.cos(theta / 2),
            0,
            0,
            1j * pc.sin(theta / 2),
            0,
            pc.cos(theta / 2),
            -1j * pc.sin(theta / 2),
            0,
            0,
            -1j * pc.sin(theta / 2),
            pc.cos(theta / 2),
            0,
            1j * pc.sin(theta / 2),
            0,
            0,
            pc.cos(theta / 2),
        ],
        dtype=state.cp_type,
    )
    _apply_two_qubit_matrix(state, qubits, matrix)


def RZZ(
    state: CuStateVec,
    qubits: tuple[int, int],
    angles: tuple[float],
    **_params: SimulatorGateParams,
) -> None:
    """Apply a rotation about ZZ.

    Args:
        state: An instance of CuStateVec
        qubits: A tuple with the index of the qubits where the gate is applied
        angles: A tuple containing a single angle in radians
    """
    if len(angles) != 1:
        msg = "Gate must be given 1 angle parameter."
        raise ValueError(msg)
    theta = angles[0]

    matrix = cp.asarray(
        [
            np.exp(-1j * theta / 2),
            0,
            0,
            0,
            0,
            np.exp(1j * theta / 2),
            0,
            0,
            0,
            0,
            np.exp(1j * theta / 2),
            0,
            0,
            0,
            0,
            np.exp(-1j * theta / 2),
        ],
        dtype=state.cp_type,
    )
    _apply_two_qubit_matrix(state, qubits, matrix)


def R2XXYYZZ(
    state: CuStateVec,
    qubits: tuple[int, int],
    angles: tuple[float, float, float],
    **_params: SimulatorGateParams,
) -> None:
    """Apply RXX*RYY*RZZ.

    Args:
        state: An instance of CuStateVec
        qubits: A tuple with the index of the qubits where the gate is applied
        angles: A tuple containing three angles in radians, for XX, YY and ZZ, in that order
    """
    if len(angles) != 3:
        msg = "Gate must be given 3 angle parameters."
        raise ValueError(msg)

    RXX(state, qubits, (angles[0],))
    RYY(state, qubits, (angles[1],))
    RZZ(state, qubits, (angles[2],))


def SXX(
    state: CuStateVec,
    qubits: tuple[int, int],
    **_params: SimulatorGateParams,
) -> None:
    """Apply a square root of XX gate.

    Args:
        state: An instance of CuStateVec
        qubits: A tuple with the index of the qubits where the gate is applied
    """
    RXX(state, qubits, angles=(pc.f64.frac_pi_2,))


def SXXdg(
    state: CuStateVec,
    qubits: tuple[int, int],
    **_params: SimulatorGateParams,
) -> None:
    """Apply adjoint of a square root of XX gate.

    Args:
        state: An instance of CuStateVec
        qubits: A tuple with the index of the qubits where the gate is applied
    """
    RXX(state, qubits, angles=(-pc.f64.frac_pi_2,))


def SYY(
    state: CuStateVec,
    qubits: tuple[int, int],
    **_params: SimulatorGateParams,
) -> None:
    """Apply a square root of YY gate.

    Args:
        state: An instance of CuStateVec
        qubits: A tuple with the index of the qubits where the gate is applied
    """
    RYY(state, qubits, angles=(pc.f64.frac_pi_2,))


def SYYdg(
    state: CuStateVec,
    qubits: tuple[int, int],
    **_params: SimulatorGateParams,
) -> None:
    """Apply adjoint of a square root of YY gate.

    Args:
        state: An instance of CuStateVec
        qubits: A tuple with the index of the qubits where the gate is applied
    """
    RYY(state, qubits, angles=(-pc.f64.frac_pi_2,))


def SZZ(
    state: CuStateVec,
    qubits: tuple[int, int],
    **_params: SimulatorGateParams,
) -> None:
    """Apply a square root of ZZ gate.

    Args:
        state: An instance of CuStateVec
        qubits: A tuple with the index of the qubits where the gate is applied
    """
    RZZ(state, qubits, angles=(pc.f64.frac_pi_2,))


def SZZdg(
    state: CuStateVec,
    qubits: tuple[int, int],
    **_params: SimulatorGateParams,
) -> None:
    """Apply adjoint of a square root of ZZ gate.

    Args:
        state: An instance of CuStateVec
        qubits: A tuple with the index of the qubits where the gate is applied
    """
    RZZ(state, qubits, angles=(-pc.f64.frac_pi_2,))


def SWAP(
    state: CuStateVec,
    qubits: tuple[int, int],
    **_params: SimulatorGateParams,
) -> None:
    """Apply a SWAP gate.

    Args:
        state: An instance of CuStateVec
        qubits: A tuple with the index of the qubits where the gate is applied
    """
    if qubits[0] >= state.num_qubits or qubits[0] < 0:
        msg = f"Qubit {qubits[0]} out of range."
        raise ValueError(msg)
    if qubits[1] >= state.num_qubits or qubits[1] < 0:
        msg = f"Qubit {qubits[1]} out of range."
        raise ValueError(msg)
    # CuStateVec uses smaller qubit index as least significant
    q0 = state.num_qubits - 1 - qubits[0]
    q1 = state.num_qubits - 1 - qubits[1]

    # Possibly faster since it may just be an internal qubit relabelling or sv reshape
    cusv.apply_generalized_permutation_matrix(
        handle=state.libhandle,
        sv=state.cupy_vector.data.ptr,
        sv_data_type=state.cuda_type,
        n_index_bits=state.num_qubits,
        permutation=[
            0,
            2,
            1,
            3,
        ],  # Leave |00> and |11> where they are, swap the other two
        diagonals=0,  # Don't apply a diagonal gate
        diagonals_data_type=state.cuda_type,
        adjoint=0,  # Don't use the adjoint
        targets=[q0, q1],
        n_targets=2,
        controls=[],
        control_bit_values=[],  # No value of control bit assigned
        n_controls=0,
        extra_workspace=0,  # Let cuQuantum use the mempool we configured
        extra_workspace_size_in_bytes=0,  # Let cuQuantum use the mempool we configured
    )
    state.stream.synchronize()


def G(
    state: CuStateVec,
    qubits: tuple[int, int],
    **_params: SimulatorGateParams,
) -> None:
    """'G': (('I', 'H'), 'CNOT', ('H', 'H'), 'CNOT', ('I', 'H'))."""
    H(state, qubits[1])
    CX(state, qubits)
    H(state, qubits[0])
    H(state, qubits[1])
    CX(state, qubits)
    H(state, qubits[1])



# Copyright 2025 The PECOS Developers
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance with
# the License.You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
# specific language governing permissions and limitations under the License.

"""QASM regression test configuration and fixtures."""

from __future__ import annotations

from pathlib import Path
from typing import TYPE_CHECKING

import pytest
from pecos.slr import SlrConverter

if TYPE_CHECKING:
    from collections.abc import Callable


@pytest.fixture
def compare_qasm() -> Callable[..., None]:
    """Pytest fixture for comparing QASM output against regression files."""

    def _compare_qasm(
        block: object,
        *params: object,
        directory: Path | None = None,
        filename: str | None = None,
    ) -> None:
        if directory is None:
            directory = Path(__file__).parent

        if filename is None:
            filename = str(type(block))[8:-2]

        if params:
            params = [str(p) for p in params]
            params = "_".join(params)
            filename = f"{filename}_{params}"

        filename = f"{filename}.qasm"
        file_dir = directory / "regression_qasm" / filename

        with Path(file_dir).open(encoding="utf-8") as file:
            qasm1 = file.read()

        qasm1 = qasm1.strip()
        # TODO: Fix this... this is kinda hacky
        if hasattr(block, "qargs") and hasattr(block, "params") and hasattr(block, "sym"):
            qasm2 = block.gen("qasm").strip()
        elif hasattr(block, "gen"):
            qasm2 = block.gen("qasm", add_versions=False).strip()
        else:
            qasm2 = SlrConverter(block).qasm(add_versions=False).strip()

        assert qasm1 == qasm2

    return _compare_qasm



# Copyright 2018 The PECOS Developers
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance with
# the License.You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
# specific language governing permissions and limitations under the License.

"""Single-qubit gate operations for Pauli fault propagation simulator.

This module provides single-qubit quantum gate operations for the Pauli fault propagation simulator, including
Clifford gates and their effect on Pauli frame propagation for efficient stabilizer circuit simulation.
"""

from __future__ import annotations

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from pecos.simulators.pauliprop.state import PauliProp
    from pecos.typing import SimulatorGateParams


def switch(
    state: PauliProp,
    switch_list: list[tuple[str, str]],
    qubit: int,
) -> None:
    """Switch Pauli fault type on a qubit according to provided mapping.

    Transforms the fault type on a qubit by checking the current fault
    against a list of transformation rules and applying the first match.

    Args:
        state: Pauli fault propagation state containing fault sets.
        switch_list: List of (initial_fault, final_fault) transformation pairs.
        qubit: Index of the qubit to apply transformation to.
    """
    for symbol_init, symbol_final in switch_list:
        if qubit in state.faults[symbol_init]:
            state.faults[symbol_init].remove(qubit)
            state.faults[symbol_final].add(qubit)
            break


def Identity(state: PauliProp, qubit: int, **_params: SimulatorGateParams) -> None:
    """Identity does nothing.

    X -> X
    Z -> Z
    Y -> Y

    Args:
        state (PauliProp): The class representing the Pauli fault state.
        qubit (int): An integer indexing the qubit being operated on.

    Returns: None

    """


def X(state: PauliProp, qubit: int, **_params: SimulatorGateParams) -> None:
    """Pauli X.

    X -> X
    Z -> -Z
    Y -> -Y

    Args:
        state (PauliProp): The class representing the Pauli fault state.
        qubit (int): An integer indexing the qubit being operated on.

    Returns: None

    """
    if state.track_sign and (qubit in state.faults["Z"] or qubit in state.faults["Y"]):
        state.flip_sign()


def Y(state: PauliProp, qubit: int, **_params: SimulatorGateParams) -> None:
    """Apply Pauli Y gate.

    X -> -X
    Z -> -Z
    Y -> Y.

    Args:
        state (PauliProp): The class representing the Pauli fault state.
        qubit (int): An integer indexing the qubit being operated on.

    Returns: None

    """
    if state.track_sign and (qubit in state.faults["X"] or qubit in state.faults["Z"]):
        state.flip_sign()


def Z(state: PauliProp, qubit: int, **_params: SimulatorGateParams) -> None:
    """Apply Pauli Z gate.

    X -> -X
    Z -> Z
    Y -> -Y.

    Args:
        state (PauliProp): The class representing the Pauli fault state.
        qubit (int): An integer indexing the qubit being operated on.

    Returns: None

    """
    if state.track_sign and (qubit in state.faults["X"] or qubit in state.faults["Y"]):
        state.flip_sign()


def SX(state: PauliProp, qubit: int, **_params: SimulatorGateParams) -> None:
    """Square root of X.

    X -> X
    Z -> -Y
    Y -> Z

    Args:
        state (PauliProp): The class representing the Pauli fault state.
        qubit (int): An integer indexing the qubit being operated on.

    Returns: None

    """
    if state.track_sign and qubit in state.faults["Z"]:
        state.flip_sign()

    switch(
        state,
        [
            ("Z", "Y"),
            ("Y", "Z"),
        ],
        qubit,
    )


def SXdg(state: PauliProp, qubit: int, **_params: SimulatorGateParams) -> None:
    """Hermitian conjugate of the square root of X.

    X -> X
    Z -> Y
    Y -> -Z

    Args:
        state (PauliProp): The class representing the Pauli fault state.
        qubit (int): An integer indexing the qubit being operated on.

    Returns: None

    """
    if state.track_sign and qubit in state.faults["Y"]:
        state.flip_sign()

    switch(
        state,
        [
            ("Z", "Y"),
            ("Y", "Z"),
        ],
        qubit,
    )


def SY(state: PauliProp, qubit: int, **_params: SimulatorGateParams) -> None:
    """Square root of Y.

    X -> -Z
    Z -> X
    Y -> Y

    Args:
        state (PauliProp): The class representing the Pauli fault state.
        qubit (int): An integer indexing the qubit being operated on.

    Returns: None

    """
    if state.track_sign and qubit in state.faults["X"]:
        state.flip_sign()

    switch(
        state,
        [
            ("X", "Z"),
            ("Z", "X"),
        ],
        qubit,
    )


def SYdg(state: PauliProp, qubit: int, **_params: SimulatorGateParams) -> None:
    """Hermitian conjugate of the square root of Y.

    X -> Z
    Z -> -X
    Y -> Y

    Args:
        state (PauliProp): The class representing the Pauli fault state.
        qubit (int): An integer indexing the qubit being operated on.

    Returns: None

    """
    if state.track_sign and qubit in state.faults["Z"]:
        state.flip_sign()

    switch(
        state,
        [
            ("X", "Z"),
            ("Z", "X"),
        ],
        qubit,
    )


def SZ(state: PauliProp, qubit: int, **_params: SimulatorGateParams) -> None:
    """Square root of Z.

    X -> Y
    Z -> Z
    Y -> -X

    Args:
        state (PauliProp): The class representing the Pauli fault state.
        qubit (int): An integer indexing the qubit being operated on.

    Returns: None

    """
    if state.track_sign and qubit in state.faults["Y"]:
        state.flip_sign()

    switch(
        state,
        [
            ("X", "Y"),
            ("Y", "X"),
        ],
        qubit,
    )


def SZdg(state: PauliProp, qubit: int, **_params: SimulatorGateParams) -> None:
    """Hermitian conjugate of the square root of Z.

    X -> -Y
    Z -> Z
    Y -> X

    Args:
        state (PauliProp): The class representing the Pauli fault state.
        qubit (int): An integer indexing the qubit being operated on.

    Returns: None

    """
    if state.track_sign and qubit in state.faults["X"]:
        state.flip_sign()

    switch(
        state,
        [
            ("X", "Y"),
            ("Y", "X"),
        ],
        qubit,
    )


def H(state: PauliProp, qubit: int, **_params: SimulatorGateParams) -> None:
    """Hadamard gate.

    X -> Z
    Z -> X
    Y -> -Y

    Args:
        state (PauliProp): The class representing the Pauli fault state.
        qubit (int): An integer indexing the qubit being operated on.

    Returns: None

    """
    if state.track_sign and qubit in state.faults["Y"]:
        state.flip_sign()

    switch(
        state,
        [
            ("X", "Z"),
            ("Z", "X"),
        ],
        qubit,
    )


def H2(state: PauliProp, qubit: int, **_params: SimulatorGateParams) -> None:
    """Hadamard-like rotation.

    X -> -Z
    Z -> -X
    Y -> -Y

    Args:
        state (PauliProp): The class representing the Pauli fault state.
        qubit (int): An integer indexing the qubit being operated on.

    Returns: None

    """
    if state.track_sign:
        state.flip_sign()

    switch(
        state,
        [
            ("X", "Z"),
            ("Z", "X"),
        ],
        qubit,
    )


def H3(state: PauliProp, qubit: int, **_params: SimulatorGateParams) -> None:
    """Hadamard-like rotation.

    X -> Y
    Z -> -Z
    Y -> X

    Args:
        state (PauliProp): The class representing the Pauli fault state.
        qubit (int): An integer indexing the qubit being operated on.

    Returns: None

    """
    if state.track_sign and qubit in state.faults["Z"]:
        state.flip_sign()

    switch(
        state,
        [
            ("X", "Y"),
            ("Y", "X"),
        ],
        qubit,
    )


def H4(state: PauliProp, qubit: int, **_params: SimulatorGateParams) -> None:
    """Hadamard-like rotation.

    X -> -Y
    Z -> -Z
    Y -> -X

    Args:
        state (PauliProp): The class representing the Pauli fault state.
        qubit (int): An integer indexing the qubit being operated on.

    Returns: None

    """
    if state.track_sign:
        state.flip_sign()

    switch(
        state,
        [
            ("X", "Y"),
            ("Y", "X"),
        ],
        qubit,
    )


def H5(state: PauliProp, qubit: int, **_params: SimulatorGateParams) -> None:
    """Hadamard-like rotation.

    X -> -X
    Z -> Y
    Y -> Z

    Args:
        state (PauliProp): The class representing the Pauli fault state.
        qubit (int): An integer indexing the qubit being operated on.

    Returns: None

    """
    if state.track_sign and qubit in state.faults["X"]:
        state.flip_sign()

    switch(
        state,
        [
            ("Z", "Y"),
            ("Y", "Z"),
        ],
        qubit,
    )


def H6(state: PauliProp, qubit: int, **_params: SimulatorGateParams) -> None:
    """Hadamard-like rotation.

    X -> -X
    Z -> -Y
    Y -> -Z

    Args:
        state (PauliProp): The class representing the Pauli fault state.
        qubit (int): An integer indexing the qubit being operated on.

    Returns: None

    """
    if state.track_sign:
        state.flip_sign()

    switch(
        state,
        [
            ("Z", "Y"),
            ("Y", "Z"),
        ],
        qubit,
    )


def F(state: PauliProp, qubit: int, **_params: SimulatorGateParams) -> None:
    """Face rotation.

    X -> Y
    Z -> X
    Y -> Z

    Args:
        state (PauliProp): The class representing the Pauli fault state.
        qubit (int): An integer indexing the qubit being operated on.

    Returns: None

    """
    switch(
        state,
        [
            ("X", "Y"),
            ("Z", "X"),
            ("Y", "Z"),
        ],
        qubit,
    )


def F2(state: PauliProp, qubit: int, **_params: SimulatorGateParams) -> None:
    """Face rotation.

    X -> -Z
    Z -> Y
    Y -> -X

    Args:
        state (PauliProp): The class representing the Pauli fault state.
        qubit (int): An integer indexing the qubit being operated on.

    Returns: None

    """
    if state.track_sign and (qubit in state.faults["X"] or qubit in state.faults["Y"]):
        state.flip_sign()

    switch(
        state,
        [
            ("X", "Z"),
            ("Z", "Y"),
            ("Y", "X"),
        ],
        qubit,
    )


def F3(state: PauliProp, qubit: int, **_params: SimulatorGateParams) -> None:
    """Face rotation.

    X -> Y
    Z -> -X
    Y -> -Z

    Args:
        state (PauliProp): The class representing the Pauli fault state.
        qubit (int): An integer indexing the qubit being operated on.

    Returns: None

    """
    if state.track_sign and (qubit in state.faults["Z"] or qubit in state.faults["Y"]):
        state.flip_sign()

    switch(
        state,
        [
            ("X", "Y"),
            ("Z", "X"),
            ("Y", "Z"),
        ],
        qubit,
    )


def F4(state: PauliProp, qubit: int, **_params: SimulatorGateParams) -> None:
    """Face rotation.

    X -> Z
    Z -> -Y
    Y -> -X

    Args:
        state (PauliProp): The class representing the Pauli fault state.
        qubit (int): An integer indexing the qubit being operated on.

    Returns: None

    """
    if state.track_sign and (qubit in state.faults["Z"] or qubit in state.faults["Y"]):
        state.flip_sign()

    switch(
        state,
        [
            ("X", "Z"),
            ("Z", "Y"),
            ("Y", "X"),
        ],
        qubit,
    )


def Fdg(state: PauliProp, qubit: int, **_params: SimulatorGateParams) -> None:
    """Face rotation.

    X -> Z
    Z -> Y
    Y -> X

    Args:
        state (PauliProp): The class representing the Pauli fault state.
        qubit (int): An integer indexing the qubit being operated on.

    Returns: None

    """
    switch(
        state,
        [
            ("X", "Z"),
            ("Z", "Y"),
            ("Y", "X"),
        ],
        qubit,
    )


def F2dg(state: PauliProp, qubit: int, **_params: SimulatorGateParams) -> None:
    """Face rotation.

    X -> -Y
    Z -> -X
    Y -> Z

    Args:
        state (PauliProp): The class representing the Pauli fault state.
        qubit (int): An integer indexing the qubit being operated on.

    Returns: None

    """
    if state.track_sign and (qubit in state.faults["X"] or qubit in state.faults["Z"]):
        state.flip_sign()

    switch(
        state,
        [
            ("X", "Y"),
            ("Z", "X"),
            ("Y", "Z"),
        ],
        qubit,
    )


def F3dg(state: PauliProp, qubit: int, **_params: SimulatorGateParams) -> None:
    """Face rotation.

    X -> -Z
    Z -> -Y
    Y -> X

    Args:
        state (PauliProp): The class representing the Pauli fault state.
        qubit (int): An integer indexing the qubit being operated on.

    Returns: None

    """
    if state.track_sign and (qubit in state.faults["X"] or qubit in state.faults["Z"]):
        state.flip_sign()

    switch(
        state,
        [
            ("X", "Z"),
            ("Z", "Y"),
            ("Y", "X"),
        ],
        qubit,
    )


def F4dg(state: PauliProp, qubit: int, **_params: SimulatorGateParams) -> None:
    """Face rotation.

    X -> -Y
    Z -> X
    Y -> -Z

    Args:
        state (PauliProp): The class representing the Pauli fault state.
        qubit (int): An integer indexing the qubit being operated on.

    Returns: None

    """
    if state.track_sign and (qubit in state.faults["X"] or qubit in state.faults["Y"]):
        state.flip_sign()

    switch(
        state,
        [
            ("X", "Y"),
            ("Z", "X"),
            ("Y", "Z"),
        ],
        qubit,
    )



"""Steane 7-qubit quantum error correction code implementation.

This module provides the main Steane class that implements the Steane 7-qubit quantum error correction code, including
all necessary operations for fault-tolerant quantum computation.
"""

# Copyright 2024 The PECOS Developers
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance with
# the License.You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
# specific language governing permissions and limitations under the License.

from __future__ import annotations

from typing import TYPE_CHECKING, NoReturn
from warnings import warn

from pecos.slr import Block, CReg, If, Permute, QReg, Vars
from pecos.slr.qeclib.steane.decoders.lookup import (
    FlagLookupQASMActiveCorrectionX,
    FlagLookupQASMActiveCorrectionZ,
)
from pecos.slr.qeclib.steane.gates_sq import paulis, sqrt_paulis
from pecos.slr.qeclib.steane.gates_sq.hadamards import H
from pecos.slr.qeclib.steane.gates_tq import transversal_tq
from pecos.slr.qeclib.steane.meas.destructive_meas import MeasDecode
from pecos.slr.qeclib.steane.preps.pauli_states import PrepRUS
from pecos.slr.qeclib.steane.preps.t_plus_state import (
    PrepEncodeTPlusFTRUS,
    PrepEncodeTPlusNonFT,
)
from pecos.slr.qeclib.steane.qec.qec_3parallel import (
    ParallelFlagQEC,
    ParallelFlagQECActiveCorrection,
)
from pecos.slr.qeclib.steane.syn_extract.bare import SynExtractBare
from pecos.slr.qeclib.steane.syn_extract.flagged import SynExtractFlagged

if TYPE_CHECKING:
    from pecos.slr import Bit


class Steane(Vars):
    """A generic implementation of a Steane code and operations.

    This represents one particular choice of Steane protocols. For finer control construct your own class
    or utilize the library of Steane code protocols directly.
    """

    def __init__(
        self,
        name: str,
        default_rus_limit: int = 3,
        ancillas: QReg | None = None,
        flag_qubits: QReg | None = None,
    ) -> None:
        """Initialize a Steane code instance with associated quantum and classical registers.

        Args:
            name: Name prefix for all registers associated with this Steane code instance.
            default_rus_limit: Default limit for repeat-until-success procedures. Defaults to 3.
            ancillas: Optional pre-existing ancilla register. If None, creates a new 3-qubit
                ancilla register. Must have at least 3 qubits if provided.
            flag_qubits: Optional pre-existing flag qubit register. If None, creates a new 3-qubit
                flag register when needed.

        Raises:
            ValueError: If provided ancilla register has fewer than 3 qubits.
        """
        super().__init__()
        # Set the source class for code generation
        self.source_class = self.__class__.__name__
        self.check_indices = [[2, 1, 3, 0], [5, 2, 1, 4], [6, 5, 2, 3]]

        self.d = QReg(f"{name}_d", 7)
        self.a = ancillas or QReg(f"{name}_a", 3)
        if flag_qubits is not None:
            self.f = flag_qubits or QReg(f"{name}_f", 3)
        else:
            self.f = None
        self.c = CReg(f"{name}_c", 32)

        # if self.a.size < 3:
        #     msg = f"Steane ancilla registers must have >= 3 qubits (provided: {self.a.size})"
        #     raise ValueError(msg)

        # TODO: Make it so I can put these in self.c... need to convert things like if(c) and c = a ^ b, a = 0;
        #  to allow lists of bits
        self.syn_meas = CReg(f"{name}_syn_meas", 32)
        self.last_raw_syn_x = CReg(f"{name}_last_raw_syn_x", 32)
        self.last_raw_syn_z = CReg(f"{name}_last_raw_syn_z", 32)
        self.scratch = CReg(f"{name}_scratch", 32)
        self.flag_x = CReg(f"{name}_flag_x", 3)
        self.flag_z = CReg(f"{name}_flags_z", 3)

        self.flags = CReg(f"{name}_flags", 3)  # weird error when using [c, c, c]

        self.raw_meas = CReg(f"{name}_raw_meas", 7)

        self.syn_x = CReg(f"{name}_syn_x", 3)
        self.syn_z = CReg(f"{name}_syn_z", 3)
        self.syndromes = CReg(f"{name}_syndromes", 3)
        self.verify_prep = CReg(f"{name}_verify_prep", 32)

        self.vars = [
            self.d,
        ]

        if ancillas is None:
            self.vars.append(self.a)

        if flag_qubits is None:
            self.vars.append(self.f)

        self.vars.extend(
            [
                self.c,
                self.syn_meas,
                self.last_raw_syn_x,
                self.last_raw_syn_z,
                self.scratch,
                self.flag_x,
                self.flag_z,
                self.flags,
                self.raw_meas,
                self.syn_x,
                self.syn_z,
                self.syndromes,
                self.verify_prep,
            ],
        )

        # derived classical registers
        c = self.c
        self.log_raw = c[1]
        self.log = c[2]
        self.pf_x = c[3]
        self.pf_z = c[4]
        self.t_meas = c[5]
        self.tdg_meas = c[6]

        self.default_rus_limit = default_rus_limit

    def p(
        self,
        state: str,
        reject: Bit | None = None,
        rus_limit: int | None = None,
    ) -> Block:
        """Prepare a logical qubit in a logical Pauli basis state."""
        block = PrepRUS(
            q=self.d,
            a=self.a[0],
            init=self.verify_prep[0],
            limit=rus_limit or self.default_rus_limit,
            state=state,
            first_round_reset=True,
        )
        if reject is not None:
            block.extend(reject.set(self.verify_prep[0]))
        return block

    def px(self, reject: Bit | None = None, rus_limit: int | None = None) -> Block:
        """Prepare logical |+X>, a.k.a. |+>."""
        return self.p("+X", reject=reject, rus_limit=rus_limit)

    def pnx(self, reject: Bit | None = None, rus_limit: int | None = None) -> Block:
        """Prepare logical |-X>, a.k.a. |->."""
        return self.p("-X", reject=reject, rus_limit=rus_limit)

    def py(self, reject: Bit | None = None, rus_limit: int | None = None) -> Block:
        """Prepare logical |+Y>, a.k.a. |+i>."""
        return self.p("+Y", reject=reject, rus_limit=rus_limit)

    def pny(self, reject: Bit | None = None, rus_limit: int | None = None) -> Block:
        """Prepare logical |-Y>, a.k.a. |-i>."""
        return self.p("-Y", reject=reject, rus_limit=rus_limit)

    def pz(self, reject: Bit | None = None, rus_limit: int | None = None) -> Block:
        """Prepare logical |+Z>, a.k.a. |0>."""
        return self.p("+Z", reject=reject, rus_limit=rus_limit)

    def pnz(self, reject: Bit | None = None, rus_limit: int | None = None) -> Block:
        """Prepare logical |-Z>, a.k.a. |1>."""
        return self.p("-Z", reject=reject, rus_limit=rus_limit)

    def nonft_prep_t_plus_state(self) -> Block:
        """Prepare logical T|+X> in a non-fault tolerant manner."""
        return PrepEncodeTPlusNonFT(
            q=self.d,
        )

    def prep_t_plus_state(
        self,
        reject: Bit | None = None,
        rus_limit: int | None = None,
    ) -> Block:
        """Prepare logical T|+X> in a fault tolerant manner."""
        block = Block(
            self.scratch.set(0),
            PrepEncodeTPlusFTRUS(
                d=self.d,
                a=self.a,
                out=self.scratch,
                reject=self.scratch[2],  # the first two bits are used by "out"
                flag_x=self.flag_x,
                flag_z=self.flag_z,
                flags=self.flags,
                last_raw_syn_x=self.last_raw_syn_x,
                last_raw_syn_z=self.last_raw_syn_z,
                limit=rus_limit or self.default_rus_limit,
            ),
        )
        if reject is not None:
            block.extend(reject.set(self.scratch[2]))
        return block

    def nonft_prep_tdg_plus_state(self) -> Block:
        """Prepare logical Tdg|+X> in a non-fault tolerant manner."""
        return Block(
            self.nonft_prep_t_plus_state(),
            self.z(),
        )

    def prep_tdg_plus_state(
        self,
        reject: Bit | None = None,
        rus_limit: int | None = None,
    ) -> Block:
        """Prepare logical Tdg|+X> in a fault tolerant manner."""
        return Block(
            self.prep_t_plus_state(reject=reject, rus_limit=rus_limit),
            self.szdg(),
        )

    def x(self) -> Block:
        """Logical Pauli X gate."""
        return paulis.X(self.d)

    def y(self) -> Block:
        """Logical Pauli Y gate."""
        return paulis.Y(self.d)

    def z(self) -> Block:
        """Logical Pauli Z gate."""
        return paulis.Z(self.d)

    def h(self) -> Block:
        """Logical Hadamard gate."""
        return H(self.d)

    def sx(self) -> Block:
        """Sqrt of X."""
        return sqrt_paulis.SX(self.d)

    def sxdg(self) -> Block:
        """Adjoint of sqrt of X."""
        return sqrt_paulis.SXdg(self.d)

    def sy(self) -> Block:
        """Sqrt of Y."""
        return sqrt_paulis.SY(self.d)

    def sydg(self) -> Block:
        """Adjoint of sqrt of Y."""
        return sqrt_paulis.SYdg(self.d)

    def sz(self) -> Block:
        """Sqrt of Z. Also known as the S gate."""
        return sqrt_paulis.SZ(self.d)

    def nonft_t(self, aux: Steane) -> Block:
        """T gate via teleportation using non-fault-tolerant initialization of the T|+> state."""
        return Block(
            aux.nonft_prep_t_plus_state(),
            self.cx(aux),
            aux.mz(self.t_meas),
            If(self.t_meas == 1).Then(self.sz()),
        )

    def t(
        self,
        aux: Steane,
        reject: Bit | None = None,
        rus_limit: int | None = None,
    ) -> Block:
        """T gate via teleportation using fault-tolerant initialization of the T|+> state."""
        return Block(
            aux.prep_t_plus_state(reject=reject, rus_limit=rus_limit),
            self.cx(aux),
            aux.mz(self.t_meas),
            If(self.t_meas == 1).Then(self.sz()),  # SZ/S correction.
        )

    def nonft_tdg(self, aux: Steane) -> Block:
        """Tdg gate via teleportation using non-fault-tolerant initialization of the Tdg|+> state."""
        return Block(
            aux.nonft_prep_tdg_plus_state(),
            self.cx(aux),
            aux.mz(self.tdg_meas),
            If(self.tdg_meas == 1).Then(self.szdg()),
        )

    def tdg(
        self,
        aux: Steane,
        reject: Bit | None = None,
        rus_limit: int | None = None,
    ) -> Block:
        """Tdg gate via teleportation using fault-tolerant initialization of the Tdg|+> state."""
        return Block(
            aux.prep_tdg_plus_state(reject=reject, rus_limit=rus_limit),
            self.cx(aux),
            aux.mz(self.tdg_meas),
            If(self.tdg_meas == 1).Then(self.szdg()),  # SZdg/Sdg correction.
        )

    #  Begin Experimental: ------------------------------------

    def nonft_t_tel(self, aux: Steane) -> Block:
        """T gate via teleportation (non-fault-tolerant, experimental).

        Warning:
            This is experimental.

        T gate via teleportation using non-fault-tolerant initialization of the T|+> state.

        This version teleports the logical qubit from the original qubit to the auxiliary logical qubit. For
        convenience, the qubits are relabeled, so you can continue to use the original Steane code logical qubit.
        """
        warn("Using experimental feature: nonft_t_tel", stacklevel=2)
        return Block(
            aux.nonft_prep_t_plus_state(),
            aux.cx(self),
            self.mz(self.t_meas),
            If(self.t_meas == 1).Then(aux.x(), aux.sz()),
            self.permute(aux),
        )

    def t_tel(
        self,
        aux: Steane,
        reject: Bit | None = None,
        rus_limit: int | None = None,
    ) -> Block:
        """T gate via teleportation (fault-tolerant, experimental).

        Warning:
            This is experimental.

        T gate via teleportation using fault-tolerant initialization of the T|+> state.

        This version teleports the logical qubit from the original qubit to the auxiliary logical qubit. For
        convenience, the qubits are relabeled, so you can continue to use the original Steane code logical qubit.
        """
        warn("Using experimental feature: t_tel", stacklevel=2)
        return Block(
            aux.prep_t_plus_state(reject=reject, rus_limit=rus_limit),
            aux.cx(self),
            self.mz(self.t_meas),
            If(self.t_meas == 1).Then(aux.x(), aux.sz()),  # SZ/S correction.
            self.permute(aux),
        )

    def nonft_tdg_tel(self, aux: Steane) -> Block:
        """T† gate via teleportation (non-fault-tolerant, experimental).

        Warning:
            This is experimental.

        Tdg gate via teleportation using non-fault-tolerant initialization of the Tdg|+> state.

        This version teleports the logical qubit from the original qubit to the auxiliary logical qubit. For
        convenience, the qubits are relabeled, so you can continue to use the original Steane code logical qubit.
        """
        warn("Using experimental feature: nonft_tdg_tel", stacklevel=2)
        return Block(
            aux.nonft_prep_tdg_plus_state(),
            aux.cx(self),
            self.mz(self.tdg_meas),
            If(self.tdg_meas == 1).Then(aux.x(), aux.szdg()),
            self.permute(aux),
        )

    def tdg_tel(
        self,
        aux: Steane,
        reject: Bit | None = None,
        rus_limit: int | None = None,
    ) -> Block:
        """T† gate via teleportation (fault-tolerant, experimental).

        Warning:
            This is experimental.

        Tdg gate via teleportation using fault-tolerant initialization of the Tdg|+> state.

        This version teleports the logical qubit from the original qubit to the auxiliary logical qubit. For
        convenience, the qubits are relabeled, so you can continue to use the original Steane code logical qubit.
        """
        warn("Using experimental feature: tdg_tel", stacklevel=2)
        return Block(
            aux.prep_tdg_plus_state(reject=reject, rus_limit=rus_limit),
            aux.cx(self),
            self.mz(self.tdg_meas),
            If(self.t_meas == 1).Then(aux.x(), aux.szdg()),  # SZdg/Sdg correction.
            self.permute(aux),
        )

    def t_cor(
        self,
        aux: Steane,
        reject: Bit | None = None,
        flag: Bit | None = None,
        rus_limit: int | None = None,
    ) -> Block:
        """T gate via teleportation using fault-tolerant initialization of the T|+> state.

        Applies active corrections of errors diagnozed by the measurement for gate teleportation.
        """
        warn("Using experimental feature: t_cor", stacklevel=2)
        block = Block(
            # gate teleportation without logical correction
            aux.prep_t_plus_state(reject=reject, rus_limit=rus_limit),
            self.cx(aux),
            aux.mz(self.t_meas),
            # active error correction
            self.syn_z.set(aux.syn_meas),
            self.last_raw_syn_z.set(0),
            self.pf_x.set(0),
            FlagLookupQASMActiveCorrectionZ(
                self.d,
                self.syn_z,
                self.syn_z,
                self.last_raw_syn_z,
                self.pf_x,
                self.syn_z,
                self.syn_z,
                self.scratch,
            ),
            # logical correction
            If(self.t_meas == 1).Then(self.sz()),
        )
        if flag is not None:
            block.extend(If(self.syn_z != 0).Then(flag.set(1)))
        return block

    def tdg_cor(
        self,
        aux: Steane,
        reject: Bit | None = None,
        flag: Bit | None = None,
        rus_limit: int | None = None,
    ) -> Block:
        """Tdg gate via teleportation using fault-tolerant initialization of the Tdg|+> state.

        Applies active corrections of errors diagnozed by the measurement for gate teleportation.
        """
        warn("Using experimental feature: t_cor", stacklevel=2)
        block = Block(
            # gate teleportation without logical correction
            aux.prep_tdg_plus_state(reject=reject, rus_limit=rus_limit),
            self.cx(aux),
            aux.mz(self.tdg_meas),
            # active error correction
            self.syn_z.set(aux.syn_meas),
            self.last_raw_syn_z.set(0),
            self.pf_x.set(0),
            FlagLookupQASMActiveCorrectionZ(
                self.d,
                self.syn_z,
                self.syn_z,
                self.last_raw_syn_z,
                self.pf_x,
                self.syn_z,
                self.syn_z,
                self.scratch,
            ),
            # logical correction
            If(self.tdg_meas == 1).Then(self.szdg()),
        )
        if flag is not None:
            block.extend(If(self.syn_z != 0).Then(flag.set(1)))
        return block

    # End Experimental: ------------------------------------

    def szdg(self) -> Block:
        """Adjoint of Sqrt of Z. Also known as the Sdg gate."""
        return sqrt_paulis.SZdg(self.d)

    def cx(self, target: Steane) -> Block:
        """Logical CX."""
        return transversal_tq.CX(self.d, target.d)

    def cy(self, target: Steane) -> Block:
        """Logical CY."""
        return transversal_tq.CY(self.d, target.d)

    def cz(self, target: Steane) -> Block:
        """Logical CZ."""
        return transversal_tq.CZ(self.d, target.d)

    def m(self, meas_basis: str, log: Bit | None = None) -> Block:
        """Destructively measure the logical qubit in some Pauli basis."""
        block = MeasDecode(
            q=self.d,
            meas_basis=meas_basis,
            meas=self.raw_meas,
            log_raw=self.log_raw,
            log=self.log,
            syn_meas=self.syn_meas,
            pf_x=self.pf_x,
            pf_z=self.pf_z,
            last_raw_syn_x=self.last_raw_syn_x,
            last_raw_syn_z=self.last_raw_syn_z,
        )
        if log is not None:
            block.extend(log.set(self.log))
        return block

    def mx(self, log: Bit | None = None) -> Block:
        """Logical destructive measurement of the logical X operator."""
        return self.m("X", log=log)

    def my(self, log: Bit | None = None) -> Block:
        """Logical destructive measurement of the logical Y operator."""
        return self.m("Y", log=log)

    def mz(self, log: Bit | None = None) -> Block:
        """Logical destructive measurement of the logical Z operator."""
        return self.m("Z", log=log)

    def qec(self, flag: Bit | None = None) -> Block:
        """Perform quantum error correction using parallel flag-based active correction.

        Args:
            flag: Optional flag bit for conditional execution.

        Returns:
            Block containing the quantum error correction operations.
        """
        block = ParallelFlagQECActiveCorrection(
            q=self.d,
            a=self.a,
            flag_x=self.flag_x,
            flag_z=self.flag_z,
            flags=self.flags,
            syn_x=self.syn_x,
            syn_z=self.syn_z,
            last_raw_syn_x=self.last_raw_syn_x,
            last_raw_syn_z=self.last_raw_syn_z,
            syndromes=self.syndromes,
            pf_x=self.pf_x,
            pf_z=self.pf_z,
            scratch=self.scratch,
        )
        if flag is not None:
            block.extend(If(self.flags != 0).Then(flag.set(1)))
        return block

    def qec_not_active(
        self,
        flag: Bit | None = None,
        pf_x: Bit | None = None,
        pf_z: Bit | None = None,
        flag_x: CReg | None = None,
        flag_z: CReg | None = None,
        syn_x: CReg | None = None,
        syn_z: CReg | None = None,
    ) -> Block:
        """Perform quantum error correction using parallel flag-based without active correction.

        There are potentially three syndrome extraction paths take:
            0: XZZ flag_x = 000, flag_z = 000 -> ZXX flag_x = 000, flag_z = 000 -> Done
            1: XZZ flag_x = 00*, flag_z = **0 -> measure XXXZZZ (syn_x, syn_z)
            2: XZZ flag_x = 000, flag_z = 000 -> ZXX flag_x = **0, flag_z = 00* -> measure XXXZZZ (syn_x, syn_z)
        (where at least one of the *s is 1)

        Therefore:
            if flag_x & flag_z == 0, we went down path 0
            if flag_x[0] | flag_z[1] | flag_z[2] == 1, we went down path 1
            if flag_x[1] | flag_z[2] | flag_z[0] == 1, we went down path 2

        Args:
            flag: Optional flag bit for conditional execution.
            flag_x: Optional CReg of the syndrome measured by the X checks for the first two of flagged syndrome
                   extractions. It is a raw syndrome made during the first two rounds of syndrome extraction.
            flag_z: Optional CReg of the syndrome measured by the X checks for the first two of flagged syndrome
                   extractions. It is a raw syndrome made during the first two rounds of syndrome extraction.
            syn_x: Optional CReg of the syndrome measured by the X checks for the last round of non-flagged syndrome
                   extraction. It is a raw syndrome made during the final round of syndrome extraction.
            syn_z: Optional CReg of the syndrome measured by the Z checks for the last round of non-flagged syndrome
                   extraction. It is a raw syndrome made during the final round of syndrome extraction.
            pf_x: Optional Pauli frame bit for logical X corrections determined by lookup table decoder
            pf_z: Optional Pauli frame bit for logical Z corrections determined by lookup table decoder

        Returns:
            Block containing the quantum error correction operations.
        """
        block = Block()

        block.extend(
            ParallelFlagQEC(
                q=self.d,
                a=self.a,
                flag_x=self.flag_x,
                flag_z=self.flag_z,
                flags=self.flags,
                syn_x=self.syn_x,
                syn_z=self.syn_z,
                last_raw_syn_x=self.last_raw_syn_x,
                last_raw_syn_z=self.last_raw_syn_z,
                syndromes=self.syndromes,
                pf_x=self.pf_x,
                pf_z=self.pf_z,
                scratch=self.scratch,
            ),
        )
        if flag is not None:
            block.extend(
                If(self.flags != 0).Then(flag.set(1)),
            )

        if flag_x is not None:
            if len(flag_x) != 3:
                msg = f"flag_x must have length 3, got {len(flag_x)}"
                raise ValueError(msg)
            block.extend(flag_x.set(self.flag_x))

        if flag_z is not None:
            if len(flag_z) != 3:
                msg = f"flag_z must have length 3, got {len(flag_z)}"
                raise ValueError(msg)
            block.extend(flag_z.set(self.flag_z))

        if syn_x is not None:
            if len(syn_x) != 3:
                msg = f"syn_x must have length 3, got {len(syn_x)}"
                raise ValueError(msg)
            block.extend(syn_x.set(self.syn_x))

        if syn_z is not None:
            if len(syn_z) != 3:
                msg = f"syn_z must have length 3, got {len(syn_z)}"
                raise ValueError(msg)
            block.extend(syn_z.set(self.syn_z))

        if pf_x is not None:
            block.extend(pf_x.set(self.pf_x))

        if pf_z is not None:
            block.extend(pf_z.set(self.pf_z))

        return block

    def qec_steane(
        self,
        aux: Steane,
        reject_x: Bit | None = None,
        reject_z: Bit | None = None,
        flag_x: Bit | None = None,
        flag_z: Bit | None = None,
        rus_limit: int | None = None,
    ) -> Block:
        """Run a Steane-type error-correction cycle of this code."""
        return Block(
            self.qec_steane_x(
                aux,
                reject=reject_x,
                flag=flag_x,
                rus_limit=rus_limit,
            ),
            self.qec_steane_z(
                aux,
                reject=reject_z,
                flag=flag_z,
                rus_limit=rus_limit,
            ),
        )

    def qec_steane_x(
        self,
        aux: Steane,
        reject: Bit | None = None,
        flag: Bit | None = None,
        rus_limit: int | None = None,
    ) -> Block:
        """Run a Steane-type error-correction cycle for X errors."""
        warn("Using experimental feature: qec_steane_x", stacklevel=2)
        block = Block(
            aux.px(reject=reject, rus_limit=rus_limit),
            self.cx(aux),
            aux.mz(),
            self.syn_z.set(aux.syn_meas),
            self.last_raw_syn_z.set(0),
            self.pf_x.set(0),
            FlagLookupQASMActiveCorrectionZ(
                self.d,
                self.syn_z,
                self.syn_z,
                self.last_raw_syn_z,
                self.pf_x,
                self.syn_z,
                self.syn_z,
                self.scratch,
            ),
        )
        if flag is not None:
            block.extend(If(self.syn_z != 0).Then(flag.set(1)))
        return block

    def qec_steane_z(
        self,
        aux: Steane,
        reject: Bit | None = None,
        flag: Bit | None = None,
        rus_limit: int | None = None,
    ) -> Block:
        """Run a Steane-type error-correction cycle for Z errors."""
        warn("Using experimental feature: qec_steane_z", stacklevel=2)
        block = Block(
            aux.pz(reject=reject, rus_limit=rus_limit),
            aux.cx(self),
            aux.mx(),
            self.syn_x.set(aux.syn_meas),
            self.last_raw_syn_x.set(0),
            self.pf_z.set(0),
            FlagLookupQASMActiveCorrectionX(
                self.d,
                self.syn_x,
                self.syn_x,
                self.last_raw_syn_x,
                self.pf_z,
                self.syn_x,
                self.syn_x,
                self.scratch,
            ),
        )
        if flag is not None:
            block.extend(If(self.syn_x != 0).Then(flag.set(1)))
        return block

    def qec_tel(
        self,
        aux: Steane,
        reject_x: Bit | None = None,
        reject_z: Bit | None = None,
        flag_x: Bit | None = None,
        flag_z: Bit | None = None,
        rus_limit: int | None = None,
    ) -> Block:
        """Run a teleportation-based error correction cycle."""
        return Block(
            self.qec_tel_x(aux, reject_x, flag_x, rus_limit),
            self.qec_tel_z(aux, reject_z, flag_z, rus_limit),
        )

    def qec_tel_x(
        self,
        aux: Steane,
        reject: Bit | None = None,
        flag: Bit | None = None,
        rus_limit: int | None = None,
    ) -> Block:
        """Run a teleportation-based error correction cycle for X errors."""
        warn("Using experimental feature: qec_tel_x", stacklevel=2)
        block = Block(
            # teleport
            aux.px(reject=reject, rus_limit=rus_limit),
            aux.cx(self),
            self.mz(),
            If(self.log == 1).Then(aux.x()),
            Permute(self.d, aux.d),
            # update syndromes and pauli frame
            self.last_raw_syn_x.set(0),
            self.last_raw_syn_z.set(0),
            self.syn_z.set(self.syn_meas),
            self.pf_x.set(0),
        )
        if flag is not None:
            block.extend(If(self.syn_meas != 0).Then(flag.set(1)))
        return block

    def qec_tel_z(
        self,
        aux: Steane,
        reject: Bit | None = None,
        flag: Bit | None = None,
        rus_limit: int | None = None,
    ) -> Block:
        """Run a teleportation-based error correction cycle for Z errors."""
        warn("Using experimental feature: qec_tel_z", stacklevel=2)
        block = Block(
            # teleport
            aux.pz(reject=reject, rus_limit=rus_limit),
            self.cx(aux),
            self.mx(),
            If(self.log == 1).Then(aux.z()),
            Permute(self.d, aux.d),
            # update syndromes and pauli frame
            self.last_raw_syn_x.set(0),
            self.last_raw_syn_z.set(0),
            self.syn_x.set(self.syn_meas),
            self.pf_z.set(0),
        )
        if flag is not None:
            block.extend(If(self.syn_meas != 0).Then(flag.set(1)))
        return block

    def qec_knill(self) -> NoReturn:
        """Prepare a Bell state and then teleport."""
        # TODO: ...
        msg = "qec_knill not implemented."
        raise NotImplementedError(msg)

    def syn_bare(self, syn: CReg) -> Block:
        """One single syndrome bit per check using bare syndrome extraction."""
        return SynExtractBare(self.d, self.a, self.check_indices, syn)

    def syn_flagged(self, syn: CReg, flags: CReg) -> Block:
        """One single syndrome bit and one single flag bit per check."""
        return SynExtractFlagged(self.d, self.a, self.f, self.check_indices, syn, flags)

    def syn_2para_v1_flagged(self) -> NoReturn:
        """Two-parallel syndrome extraction version 1 with flagging (not implemented).

        Raises:
            NotImplementedError: This method is not yet implemented.
        """
        # TODO: ...
        msg = "syn_2para_v1_flagged not implemented."
        raise NotImplementedError(msg)

    def syn_2para_v2_flagged(self) -> NoReturn:
        """Two-parallel syndrome extraction version 2 with flagging (not implemented).

        Raises:
            NotImplementedError: This method is not yet implemented.
        """
        # TODO: ...
        msg = "syn_2para_v2_flagged not implemented."
        raise NotImplementedError(msg)

    def permute(self, other: Steane) -> Block:
        """Permute this code block (including both quantum and classical registers) with another."""
        block = Block(
            Permute(self.d, other.d),
            Permute(self.a, other.a),
        )
        # TODO: Use Permute on classical variables rather that a custom solution
        for var_a, var_b in zip(self.vars, other.vars, strict=False):
            if isinstance(var_a, CReg):
                block.extend(
                    var_a.set(var_a ^ var_b),
                    var_b.set(var_b ^ var_a),
                    var_a.set(var_a ^ var_b),
                )
        return block



# Copyright 2025 The PECOS Developers
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance with
# the License.You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
# specific language governing permissions and limitations under the License.

"""Measurement operations for Qulacs simulator.

This module provides quantum measurement operations for the Qulacs simulator.
"""

from __future__ import annotations

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from pecos.simulators.qulacs import Qulacs
    from pecos.typing import SimulatorGateParams


def meas_z(state: Qulacs, qubit: int, **_params: SimulatorGateParams) -> int:
    """Measure qubit in Z basis.

    Args:
        state: An instance of Qulacs
        qubit: The index of the qubit to measure

    Returns:
        The measurement outcome (0 or 1)
    """
    result = state.qulacs_state.run_1q_gate("MZ", qubit)
    return int(result) if result is not None else 0



"""Test arithmetic and boolean type support in Guppy->Selene pipeline."""

import pytest
from guppylang import guppy
from guppylang.std.quantum import h, measure, qubit
from pecos import Guppy, sim
from pecos_rslib import state_vector


def test_integer_arithmetic() -> None:
    """Test integer arithmetic operations."""

    @guppy
    def quantum_add() -> bool:
        q = qubit()
        x = 3
        y = 2
        result = x + y  # result = 5

        if result > 3:  # 5 > 3, so H gate applied
            h(q)

        return measure(q)

    results = sim(Guppy(quantum_add)).qubits(1).quantum(state_vector()).seed(42).run(10).to_dict()

    raw_measurements = results.get("measurements", [])
    # For single bool return, measurements is [[1], [0], ...]
    measurements = [m[-1] if isinstance(m, list) else m for m in raw_measurements]
    assert len(measurements) == 10
    # H gate should give mix of 0s and 1s
    assert 0 in measurements
    assert 1 in measurements


def test_boolean_operations() -> None:
    """Test boolean logic operations."""

    @guppy
    def quantum_bool_logic() -> bool:
        q1 = qubit()
        q2 = qubit()
        h(q1)
        m1 = measure(q1)
        m2 = measure(q2)
        return m1 and not m2

    results = sim(Guppy(quantum_bool_logic)).qubits(2).quantum(state_vector()).seed(42).run(10).to_dict()

    raw_measurements = results.get("measurements", [])
    measurements = [m[-1] if isinstance(m, list) else m for m in raw_measurements]
    assert len(measurements) == 10


def test_integer_comparisons() -> None:
    """Test integer comparison operations."""

    @guppy
    def quantum_compare() -> bool:
        q = qubit()
        threshold = 42
        value = 50

        if value > threshold:
            h(q)

        return measure(q)

    results = sim(Guppy(quantum_compare)).qubits(1).quantum(state_vector()).seed(42).run(10).to_dict()

    raw_measurements = results.get("measurements", [])
    measurements = [m[-1] if isinstance(m, list) else m for m in raw_measurements]
    assert len(measurements) == 10
    assert 0 in measurements
    assert 1 in measurements


def test_arithmetic_in_loop() -> None:
    """Test arithmetic in loop control."""

    @guppy
    def quantum_loop() -> bool:
        q = qubit()
        count = 0
        max_count = 3

        while count < max_count:
            if count == 1:  # Only apply H on second iteration
                h(q)
            count = count + 1

        return measure(q)

    results = sim(Guppy(quantum_loop)).qubits(1).quantum(state_vector()).seed(42).run(10).to_dict()

    raw_measurements = results.get("measurements", [])
    measurements = [m[-1] if isinstance(m, list) else m for m in raw_measurements]
    assert len(measurements) == 10
    assert 0 in measurements
    assert 1 in measurements


def test_chained_comparisons() -> None:
    """Test multiple chained comparisons."""

    @guppy
    def quantum_chain() -> bool:
        q = qubit()
        a = 10
        b = 20
        c = 15

        if a < c and c < b:  # 10 < 15 < 20 is True
            h(q)

        return measure(q)

    results = sim(Guppy(quantum_chain)).qubits(1).quantum(state_vector()).seed(42).run(10).to_dict()

    raw_measurements = results.get("measurements", [])
    measurements = [m[-1] if isinstance(m, list) else m for m in raw_measurements]
    assert len(measurements) == 10
    assert 0 in measurements
    assert 1 in measurements


@pytest.mark.skip(
    reason="Conditional quantum ops based on measurement results cause register count mismatch",
)
def test_arithmetic_with_measurements() -> None:
    """Test using measurement results in arithmetic."""

    @guppy
    def quantum_measure_math() -> bool:
        q1 = qubit()
        q2 = qubit()
        h(q1)
        h(q2)

        m1 = measure(q1)
        m2 = measure(q2)

        # Use measurements in arithmetic (bools as ints)
        q3 = qubit()
        if m1 or m2:  # At least one is True
            h(q3)

        return measure(q3)

    results = sim(Guppy(quantum_measure_math)).qubits(3).quantum(state_vector()).seed(42).run(20).to_dict()

    raw_measurements = results.get("measurements", [])
    measurements = [m[-1] if isinstance(m, list) else m for m in raw_measurements]
    assert len(measurements) == 20
    # Should have mix unless both m1 and m2 are 0 (25% chance)



# Copyright 2024 The PECOS Developers
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance with
# the License.You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
# specific language governing permissions and limitations under the License.

from collections.abc import Callable
from dataclasses import dataclass
from enum import Enum
from typing import TYPE_CHECKING

import pecos as pc
from pecos.slr.qeclib import qubit as q

if TYPE_CHECKING:
    from pecos.slr.qeclib.qubit.qgate_base import QGate


@dataclass
class QG:
    qir_name: str
    adjoint: bool = False
    decomposer: Callable[["QGate"], list["QGate"]] = None

    @classmethod
    def decompose(cls, decomposer: Callable[["QGate"], list["QGate"]]):
        return cls("", adjoint=False, decomposer=decomposer)


class QIRGateMetadata(Enum):
    """Maps QEClib gates to QIR gates, including possible decompositions. See:
    https://co41-bitbucket.honeywell.lab:4443/projects/HQSSW/repos/hqcompiler/browse/hqscompiler/qir/base_validator.py#72
    """

    def __init__(self, gate: QG):
        self.qir_name = gate.qir_name
        self.decomposer = gate.decomposer

    H = QG("h")
    Y = QG("y")
    CX = QG("cnot")
    CZ = QG("cz")
    X = QG("x")
    Z = QG("z")
    RZ = QG("rz")
    RY = QG("ry")
    RX = QG("rx")
    S = QG("s")
    T = QG("t")
    R1XY = QG("u1q")
    RZZ = QG("rzz")
    SZZ = QG("zz")
    Prep = QG("reset")

    # These dagger/adjoint gates are generated with slightly different names
    Sdg = QG("s__adj")
    Tdg = QG("t__adj")

    SZ = S  # Mapped to itself in qeclib
    SZdg = Sdg  # Mapped to itself in qeclib

    # Complicated gates that require decomposition, expressed as a lambda:

    SX = QG.decompose(
        lambda sx: [
            q.RX[pc.f64.frac_pi_2](sx.qargs[0]),
        ],
    )
    SXdg = QG.decompose(
        lambda sxdg: [
            q.RX[-pc.f64.frac_pi_2](sxdg.qargs[0]),
        ],
    )
    SY = QG.decompose(
        lambda sy: [
            q.RY[pc.f64.frac_pi_2](sy.qargs[0]),
        ],
    )
    SYdg = QG.decompose(
        lambda sydg: [
            q.RY[-pc.f64.frac_pi_2](sydg.qargs[0]),
        ],
    )

    # https://github.com/PECOS-packages/PECOS/blob/development/crates/pecos-qsim/src/clifford_gateable.rs#L681
    F = QG.decompose(
        lambda f: [
            q.SX(f.qargs[0]),
            q.SZ(f.qargs[0]),
        ],
    )
    # https://github.com/PECOS-packages/PECOS/blob/development/crates/pecos-qsim/src/clifford_gateable.rs#L715
    Fdg = QG.decompose(
        lambda fdg: [
            q.SZdg(fdg.qargs[0]),
            q.SXdg(fdg.qargs[0]),
        ],
    )
    # https://github.com/PECOS-packages/PECOS/blob/development/crates/pecos-qsim/src/clifford_gateable.rs#L885
    F4 = QG.decompose(
        lambda f4: [
            q.SZ(f4.qargs[0]),
            q.SX(f4.qargs[0]),
        ],
    )
    F4dg = QG.decompose(
        lambda f4dg: [
            q.SXdg(f4dg.qargs[0]),
            # q.SZdg(f4dg.qargs[0]),
            q.RZ[-pc.f64.frac_pi_2](f4dg.qargs[0]),
        ],
    )

    # Remaining QIR gates seen:
    #     '__quantum__qis__rxxyyzz__body',



# Copyright 2024 The PECOS Developers
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance with
# the License.You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
# specific language governing permissions and limitations under the License.

"""Color 488 quantum error correction code implementation."""

from typing import Any

from pecos.qec.color.geometry import generate_488_layout


class Color488:
    """Implementation of the Color 488 quantum error correction code.

    This is a thin wrapper around pecos.qec.color geometry for use in SLR code.
    """

    def __init__(self, distance: int) -> None:
        """Initialize a Color 488 code instance.

        Args:
            distance: The code distance.
        """
        self.distance = distance
        self._layout_cache: tuple[dict[int, tuple[int, int]], list[list[Any]]] | None = None

    def get_layout(self) -> tuple[dict[int, tuple[int, int]], list[list[Any]]]:
        """Get the layout of the color 488 code.

        Returns:
            A tuple containing:
            - nodeid2pos: Mapping from node IDs to positions.
            - polygons: List of stabilizer polygons.
        """
        if self._layout_cache is None:
            self._layout_cache = generate_488_layout(self.distance)
        return self._layout_cache

    def num_data_qubits(self) -> int:
        """Get the number of data qubits in the color 488 code.

        Returns:
            int: The number of data qubits.
        """
        nodes, _ = self.get_layout()
        return len(nodes)



# Copyright 2025 The PECOS Developers
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance with
# the License.You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
# specific language governing permissions and limitations under the License.

"""QASM regression tests for qubit measurement operations."""

from collections.abc import Callable

from pecos.slr import CReg, QReg
from pecos.slr.qeclib import qubit


def test_Measure(compare_qasm: Callable[..., None]) -> None:
    """Test Measure gate QASM regression."""
    q = QReg("q_test", 1)
    m = CReg("m_test", 1)

    prog = qubit.Measure(q[0]) > m[0]
    compare_qasm(prog)



# Copyright 2018 National Technology & Engineering Solutions of Sandia, LLC (NTESS). Under the terms of Contract
# DE-NA0003525 with NTESS, the U.S. Government retains certain rights in this software.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance with
# the License.You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
# specific language governing permissions and limitations under the License.

"""A collection of useful functions."""

from typing import Any


def make_hashable_params(params: dict[str, Any]) -> tuple:
    """Checks to make sure that the parameters submitted is hashable.

    Args:
    ----
        params(dict): A dictionary of parameters to be made hashable. Values can be dicts, lists, sets, or other types.

    """
    tuple_params = []

    for key, value in params.items():
        if isinstance(value, dict):
            dict_tuple = tuple(value.items())
            tuple_params.append(dict_tuple)
        elif isinstance(value, list | set):
            tuple_params.append((key, tuple(value)))
        else:
            tuple_params.append((key, value))

    tuple_params = tuple(tuple_params)

    try:
        hash(tuple_params)
    except TypeError:
        msg = "The values of keywords given to this class must be hashable."
        raise TypeError(msg) from TypeError

    return tuple_params


def pos2qudit(layout: dict[Any, tuple]) -> dict[tuple, Any]:
    """Reverses the layout dictionary. Makes a new dictionary with (x, y, ...) => qudit_id.

    Args:
    ----
        layout: A dictionary mapping qudit IDs to their positions (coordinates).

    """
    return {p: qid for qid, p in layout.items()}


def expected_params(params: dict[str, Any], expected_set: set[str]) -> None:
    """Check that all parameters are in the expected set.

    Args:
    ----
        params(dict): Dictionary of parameters to validate
        expected_set(set): Set of expected parameter keys

    """
    keys = set(params.keys())

    unexpected = keys - expected_set
    if unexpected:
        msg = f"Received unexpected keys ({unexpected}). Expected keys include: {expected_set}"
        raise Exception(msg)
